# encoding=utf-8
# @Time    : 17-3-3
# @File    : main.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import sys
# sys.path.append('/workspace/workspace/portrait_code/tool/antgo')
# sys.path.append('/workspace/workspace/portrait_code/tool/antgo/antgo')
from antgo.ant.shell import *
from antgo.ant.generate import *
from antgo.ant.demo import *
from antgo.ant.browser import *
from antgo.ant.batch import *
from antgo.ant.activelearning import *
from antgo.ant.utils import *
from antgo.sandbox.sandbox import *
from antgo.utils.utils import *
from antgo.utils.dht import *
from antgo import version
from jinja2 import Environment, FileSystemLoader

import multiprocessing
import subprocess

if sys.version > '3':
    PY3 = True
else:
    PY3 = False


def _check_environment():
  is_in_mltalker = True if os.environ.get('ANT_ENVIRONMENT', '') != '' else False
  return is_in_mltalker


_ant_support_commands = ["train",
                         "challenge",
                         "activelearning",
                         "dataset",
                         "predict",
                         "demo",
                         "browser",
                         "startproject",
                         "config"]

#############################################
#######   antgo parameters            #######
#############################################
flags.DEFINE_string('main_file', None, 'main file')
flags.DEFINE_string('main_param', None, 'model parameters')
flags.DEFINE_string('main_folder', None, 'resource folder')
flags.DEFINE_string('version', None, 'minist antgo version')
flags.DEFINE_string('task', None, 'task file')
flags.DEFINE_string('dataset', 'images_full', 'dataset')
flags.DEFINE_string('signature', '123', 'signature')
flags.DEFINE_string('devices', '', 'devices')
flags.DEFINE_string('servers', '', '')
flags.DEFINE_string('dump', None, 'dump dir')
flags.DEFINE_string('token', None, 'token')
flags.DEFINE_string('proxy', None, 'proxy')
flags.DEFINE_string('name', None, 'name')
flags.DEFINE_string('author', None, 'author')
flags.DEFINE_string('framework', None, 'tensorflow')
flags.DEFINE_string('net', None, 'GENERAL')
flags.DEFINE_string('max_time', '100d', 'max running time')
flags.DEFINE_string('from_experiment', None, 'load model from experiment')
flags.DEFINE_string('restore_experiment', None, 'restore experiment')
flags.DEFINE_string('factory', None, '')
flags.DEFINE_string('config', None, 'config file')
flags.DEFINE_string('benchmark', None, 'benchmark experiments')
flags.DEFINE_string('host_ip', '127.0.0.1', 'host ip address')
flags.DEFINE_integer('host_port', 8909, 'port')
flags.DEFINE_string('html_template', None, 'html template')
flags.DEFINE_string('option', '', '')
flags.DEFINE_indicator('worker', '')
flags.DEFINE_indicator('master', '')
flags.DEFINE_indicator('unlabel', '')
flags.DEFINE_indicator('skip_training', '')
flags.DEFINE_string('running_platform', 'local', 'local/cloud')

FLAGS = flags.AntFLAGS
Config = config.AntConfig


def main():
  # 1.step antgo logo
  main_logo()

  # 2.step check antgo support command
  if len(sys.argv) >= 2 and \
          ((not sys.argv[1].startswith('-')) and sys.argv[1] not in _ant_support_commands):
    logger.error('antgo cli support( %s )command'%",".join(_ant_support_commands))
    sys.exit(-1)

  # 3.step parse antgo running params
  if len(sys.argv) >= 2:
    if sys.argv[1].startswith('--') or sys.argv[1].startswith('-'):
      flags.cli_param_flags(sys.argv[1:])
    else:
      flags.cli_param_flags(sys.argv[2:])

  # token
  token = FLAGS.token()
  if not PY3 and token is not None:
    token = unicode(token)

  if sys.argv[1] == 'config':
    if FLAGS.config() is not None:
      # 使用指定配置文件更新
      if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo')):
        os.makedirs(os.path.join(os.environ['HOME'], '.config', 'antgo'))

      shutil.copy(FLAGS.config(), os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml'))
      logger.info('success update config file')
    else:
      # 在当前目录生成默认配置文件
      shutil.copy(os.path.join('/'.join(os.path.realpath(__file__).split('/')[0:-1]), 'config.xml'), "./config.xml")
      logger.info('build default config file')
    return

  # 检查配置文件是否存在
  if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')):
    logger.error('missing config file, please run antgo config')
    return

  # 解析配置文件
  config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
  Config.parse_xml(config_xml)
  if len(sys.argv) == 1 or sys.argv[1].startswith('-'):
    # interactive control
    shell_process = AntShell(Context(), token)
    shell_process.start()
    return

  # 检查最低版本与当前版本的兼容性
  if FLAGS.running_platform == 'local' and FLAGS.version() is not None:
    a, b, c = FLAGS.version().split('.')
    sys_a, sys_b, sys_c = version.split('.')
    if int(sys_a) < int(a) or int(sys_b) < int(b) or int(sys_c) < int(c):
      logger.error('antgo version dont satisfy task minimum request (%s)'%FLAGS.version())
      sys.exit(-1)

  # 4.step check factory
  factory = getattr(Config, 'factory', None)
  if factory is None or \
      factory == '' or not os.path.exists(Config.factory):
    logger.error('factory folder is missing, please run antgo config')
    return

  if not os.path.exists(Config.data_factory):
    os.makedirs(Config.data_factory)
  if not os.path.exists(Config.task_factory):
    os.makedirs(Config.task_factory)

  task_factory = Config.task_factory
  data_factory = Config.data_factory

  # 5.step parse antgo execute command
  # 5.1.step other command (train, challenge, compose, deploy, server)
  ant_cmd = sys.argv[1]
  if ant_cmd not in _ant_support_commands:
    logger.error('antgo cli support( %s )command'%",".join(_ant_support_commands))
    return

  if ant_cmd == 'startproject':
    project_name = FLAGS.name()
    if project_name == "":
      project_name = 'AntgoProject'

    if os.path.exists(os.path.join(os.curdir, project_name)):
      while True:
        project_i = 0
        if not os.path.exists(os.path.join(os.curdir, '%s-%d'%(project_name, project_i))):
          project_name = '%s-%d'%(project_name, project_i)
          break

        project_i += 1

    os.makedirs(os.path.join(os.curdir, project_name))

    # generate main file and main param templates
    template_file_folder = os.path.join(os.path.dirname(__file__), 'resource', 'templates')
    file_loader = FileSystemLoader(template_file_folder)
    env = Environment(loader=file_loader)
    template = env.get_template('task_main_file.template')
    output = template.render(ModelTime=datetime.fromtimestamp(timestamp()).strftime('%Y-%m-%d'),
                             ModelName=FLAGS.name(),
                             ModelAuthor=FLAGS.author() if FLAGS.author() is not None else 'xxx',
                             tensorflow=True if FLAGS.framework() == 'tensorflow' else False,
                             GAN=True if FLAGS.net() == 'GAN' else False)

    with open(os.path.join(os.curdir, project_name, '%s_main.py' % FLAGS.name()),'w') as fp:
      fp.write(output)

    template = env.get_template('task_main_param.template')
    output = template.render(ModelTime=datetime.fromtimestamp(timestamp()).strftime('%Y-%m-%d'),
                             ModelName=FLAGS.name(),
                             ModelAuthor=FLAGS.author() if FLAGS.author() is not None else 'xxx')
    with open(os.path.join(os.curdir, project_name, '%s_param.yaml' % FLAGS.name()),'w') as fp:
      fp.write(output)

    template = env.get_template('task.template')
    output = template.render(ModelName=FLAGS.name())
    with open(os.path.join(os.curdir, project_name, '%s_task.xml' % FLAGS.name()),'w') as fp:
      fp.write(output)

    template = env.get_template('task_shell.template')
    output = ''
    if FLAGS.token() is not None:
      output = template.render(token=FLAGS.token(),
                               main_file='%s_main.py' % FLAGS.name(),
                               main_param='%s_param.yaml' % FLAGS.name())
    else:
      output = template.render(task='%s_task.xml' % FLAGS.name(),
                               main_file='%s_main.py' % FLAGS.name(),
                               main_param='%s_param.yaml' % FLAGS.name())

    with open(os.path.join(os.curdir, project_name, 'run.sh'), 'w') as fp:
      fp.write(output)

    return

  # 7.step check related params
  # 7.1 must set name
  name = FLAGS.name()
  if name is None:
    logger.error('Must set experiemnt name.')
    return

  # 7.2 check main folder (all related model code, includes main_file and main_param)
  main_folder = FLAGS.main_folder()
  if main_folder is None:
    main_folder = os.path.abspath(os.curdir)

  if os.path.exists(os.path.join(main_folder, '%s.tar.gz'%name)):
    untar_shell = 'openssl enc -d -aes256 -in %s.tar.gz -k %s | tar xz -C %s' % (name, FLAGS.signature(), '.')
    subprocess.call(untar_shell, shell=True, cwd=main_folder)
    os.remove(os.path.join(main_folder, '%s.tar.gz' % name))

    main_folder = os.path.join(main_folder, name)
    os.chdir(main_folder)

  # 7.3 check dump dir (all running data is stored here)
  dump_dir = FLAGS.dump()
  if dump_dir is None:
    dump_dir = os.path.join(os.path.abspath(os.curdir), 'dump')
    if not os.path.exists(dump_dir):
      os.makedirs(dump_dir)

  # 7.4 check main file
  main_file = FLAGS.main_file()
  if main_file is None or not os.path.exists(os.path.join(main_folder, main_file)):
    if not (FLAGS.worker() or FLAGS.master()):
      logger.error('main executing file dont exist')
      sys.exit(-1)

  # 8 step ant running
  # 8.1 step what is task
  task = FLAGS.task()
  if task is not None:
    if os.path.exists(os.path.join(task_factory, task)):
      task = os.path.join(task_factory, task)
    else:
      task = os.path.join(main_folder, task)

  dataset = FLAGS.dataset()
  if task is None and dataset is not None:
    # build default task
    with open(os.path.join(task_factory, '%s.xml'%name), 'w') as fp:
      task_content = '<task><task_name>%s</task_name>' \
                     '<input>' \
                     '<source_data><data_set_name>%s</data_set_name>' \
                     '</source_data>' \
                     '</input>' \
                     '</task>'%('default-task', dataset)
      fp.write(task_content)
    task = os.path.join(task_factory, '%s.xml'%name)

  # 8.2 step load ant context
  ant_context = None
  if main_file is not None and main_file != '':
    ant_context = main_context(main_file, main_folder)
  else:
    ant_context = Context()
  
  # 8.3 step load model config
  main_param = FLAGS.main_param()
  if main_param is not None:
    main_config_path = os.path.join(main_folder, main_param)
    try:
      with open(main_config_path, 'r') as fp:
        params = yaml.load(fp, Loader=yaml.FullLoader)
        ant_context.params = params
    except:
      print('Fail to load main_param %s.'%main_param)

  ant_context.name = name
  ant_context.data_factory = data_factory
  ant_context.task_factory = task_factory

  # 8.4 step load experiment
  if FLAGS.from_experiment() is not None and \
          FLAGS.running_platform() == 'local' and \
          ant_cmd not in ['release']:
    experiment_path = os.path.join(dump_dir, FLAGS.from_experiment())
    if not os.path.exists(experiment_path):
      logger.error('couldnt find experiment %s'%FLAGS.from_experiment())
      exit(-1)

    if os.path.exists(os.path.join(experiment_path, 'train')):
      ant_context.from_experiment = os.path.join(experiment_path, 'train')
    elif os.path.exists(os.path.join(experiment_path, 'inference')):
      ant_context.from_experiment = os.path.join(experiment_path, 'inference')
    else:
      ant_context.from_experiment = experiment_path

  # time stamp
  time_stamp = timestamp()
  if ant_cmd == "train":
    with running_sandbox(sandbox_time=FLAGS.max_time(),
                         sandbox_dump_dir=main_folder,
                         sandbox_experiment=None,
                         sandbox_user_token=token,
                         sandbox_user_proxy=FLAGS.proxy(),
                         sandbox_user_signature=FLAGS.signature()):
      running_process = AntTrain(ant_context,
                                 name,
                                 data_factory,
                                 dump_dir,
                                 token,
                                 task,
                                 main_file=main_file,
                                 main_folder=main_folder,
                                 main_param=main_param,
                                 time_stamp=time_stamp,
                                 skip_training=FLAGS.skip_training(),
                                 running_platform=FLAGS.running_platform(),
                                 proxy=FLAGS.proxy(),
                                 signature=FLAGS.signature(),
                                 devices=FLAGS.devices())
      running_process.start()
  elif ant_cmd == 'challenge':
    with running_sandbox(sandbox_dump_dir=main_folder,
                         sandbox_experiment=None,
                         sandbox_user_token=token,
                         sandbox_user_proxy=FLAGS.proxy(),
                         sandbox_user_signature=FLAGS.signature()):
      running_process = AntChallenge(ant_context,
                                     name,
                                     data_factory,
                                     dump_dir,
                                     token,
                                     task,
                                     FLAGS.benchmark(),
                                     main_file=main_file,
                                     main_folder=main_folder,
                                     main_param=main_param,
                                     time_stamp=time_stamp,
                                     running_platform=FLAGS.running_platform(),
                                     devices=FLAGS.devices())
      running_process.start()
  elif ant_cmd == "demo":
    running_process = AntDemo(ant_context,
                              name,
                              dump_dir,
                              token,
                              task,
                              html_template=FLAGS.html_template(),
                              ip=FLAGS.host_ip(),
                              port=FLAGS.host_port(),
                              time_stamp=time_stamp,
                              devices=FLAGS.devices())

    running_process.start()
  elif ant_cmd == "predict":
    running_process = AntBatch(ant_context,
                               name,
                               FLAGS.host_ip(),
                               FLAGS.host_port(),
                               token,
                               data_factory,
                               dump_dir,
                               task,
                               unlabel=FLAGS.unlabel(),
                               devices=FLAGS.devices(),
                               restore_experiment=FLAGS.restore_experiment())
    running_process.start()
  elif ant_cmd == "activelearning":
    running_process = AntActiveLearning(ant_context,
                                        name,
                                        data_factory,
                                        dump_dir,
                                        token,
                                        task,
                                        main_file=main_file,
                                        main_folder=main_folder,
                                        main_param=main_param,
                                        time_stamp=time_stamp,
                                        running_platform=FLAGS.running_platform(),
                                        max_time=FLAGS.max_time(),
                                        ip=FLAGS.host_ip(),
                                        port=FLAGS.host_port(),
                                        task=FLAGS.task(),
                                        skip_training=FLAGS.skip_training(),
                                        option=FLAGS.option(),
                                        from_experiment=FLAGS.from_experiment(),
                                        devices=FLAGS.devices())
    running_process.start()
  elif ant_cmd == 'dataset':
    running_process = AntGenerate(ant_context,
                                  name,
                                  data_factory,
                                  dump_dir,
                                  token,
                                  dataset)
    running_process.start()
  elif ant_cmd == 'browser':
    running_process = AntBrowser(ant_context,
                                 name,
                                 token,
                                 FLAGS.host_ip(),
                                 FLAGS.host_port(),
                                 data_factory,
                                 dataset,
                                 dump_dir)
    running_process.start()

  # 9.step clear context
  ant_context.wait_until_clear()

if __name__ == '__main__':
  main()



# -*- coding: UTF-8 -*-
# @Time    : 17-12-20
# @File    : layers.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import tensorflow as tf
import numpy as np


def _bilinear_filter(filter_shape, upscale_factor):
  ##filter_shape is [width, height, num_in_channels, num_out_channels]
  kernel_size = filter_shape[1]
  ### Centre location of the filter for which value is calculated
  if kernel_size % 2 == 1:
    centre_location = upscale_factor - 1
  else:
    centre_location = upscale_factor - 0.5
  
  bilinear = np.zeros([filter_shape[0], filter_shape[1]])
  for x in range(filter_shape[0]):
    for y in range(filter_shape[1]):
      ##Interpolation Calculation
      value = (1 - abs((x - centre_location) / upscale_factor)) * (1 - abs((y - centre_location) / upscale_factor))
      bilinear[x, y] = value
  weights = np.zeros(filter_shape)
  for i in range(filter_shape[2]):
    weights[:, :, i, i] = bilinear
  init = tf.constant_initializer(value=weights, dtype=tf.float32)
  
  bilinear_weights = tf.get_variable(name="filter", initializer=init, shape=weights.shape)
  return bilinear_weights


def deconv_bilinear(bottom, out_channels, in_channels, upscale_factor, output_shape=None, name=None):
  kernel_size = 2 * upscale_factor - upscale_factor % 2
  stride = upscale_factor
  strides = [1, stride, stride, 1]
  with tf.variable_scope(name, 'deconv'):
    # Shape of the bottom tensor
    in_shape = tf.shape(bottom)
    
    if output_shape is None:
      h = ((in_shape[1] - 1) * stride) + 1
      w = ((in_shape[2] - 1) * stride) + 1
      new_shape = [in_shape[0], h, w, out_channels]
      output_shape = tf.stack(new_shape)
    else:
      output_shape = tf.stack(output_shape)
    
    filter_shape = [kernel_size, kernel_size, out_channels, in_channels]

    weights = None
    if out_channels == in_channels:
      # filter with bilinear parameters
      weights = _bilinear_filter(filter_shape, upscale_factor)
    else:
      # filter with random
      weights = tf.get_variable("filter",
                                filter_shape,
                                dtype=tf.float32,
                                initializer=tf.random_normal_initializer(0, 0.02))

  return tf.nn.conv2d_transpose(bottom, weights, output_shape, strides=strides, padding='SAME')



# -*- coding: UTF-8 -*-
# Time: 10/8/17
# File: shell.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.ant.base import *
from antgo.ant import flags
from antgo.ant.challenge import *
from antgo.ant.train import *
from antgo.ant.utils import *
from antgo.resource.html import *
from antgo import config
from antgo.task.task import *
from antgo.measures import *
from antgo.measures.repeat_statistic import *
from antgo.utils import logger
from antvis.client.httprpc import *
from multiprocessing import Process
from prettytable import PrettyTable
import subprocess
import getopt
import json
import shutil
import sys
if sys.version > '3':
  PY3 = True
else:
  PY3 = False


FLAGS = flags.AntFLAGS
Config = config.AntConfig


class AntShell(AntBase):
  def __init__(self, ant_context, ant_token):
    flags.DEFINE_string('experiment_uuid', None, 'uuid')
    flags.DEFINE_string('dataset_name', None, 'dataset name')
    flags.DEFINE_string('apply_name', None, 'apply name')
    flags.DEFINE_string('task_name', None, 'task name')

    super(AntShell, self).__init__('SHELL', ant_context=ant_context, ant_token=ant_token)

  def process_task_command(self):
    task_name = FLAGS.task_name() # task name
    if task_name is None:
      response = mlogger.getEnv().dashboard.task.get()

      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      content = response['content']
      task_table = PrettyTable(["task", "time", "dataset", "apply", "publisher"])
      for task in content:
        task_table.add_row([task['task_name'],
                            time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(task['task_time'])),
                            task['task_dataset'],
                            task['task_apply_num'],
                            task['task_publisher']])

      print(task_table)
    else:
      response = mlogger.getEnv().dashboard.apply.get(apply_name=task_name)

      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      content = response['content']
      task_table = PrettyTable(["uuid", "experiment", "time", "optimum", "report", "model"])
      for experiment in content['experiments']:
        task_table.add_row([experiment['experiment_uuid'],
                            experiment['experiment_name'],
                            time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(experiment['experiment_time'])),
                            experiment['experiment_optimum'],
                            experiment['experiment_report'],
                            experiment['experiment_model']])

      print(task_table)

  def process_experiment_command(self):
    experiment_uuid = FLAGS.experiment_uuid()
    if experiment_uuid is None:
      response = mlogger.getEnv().dashboard.experiment.get()
      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      content = response['content']

      table = PrettyTable(["uuid", "task", "experiment", "time", "optimum", "report", "model"])
      for experiment in content:
        table.add_row([experiment['experiment_uuid'],
                       experiment['experiment_task'],
                       experiment['experiment_name'],
                       time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(experiment['experiment_time'])),
                       experiment['experiment_optimum'],
                       experiment['experiment_report'],
                       experiment['experiment_model']])

      print(table)
    else:
      response = mlogger.getEnv().dashboard.experiment.get(experiment_uuid=experiment_uuid)
      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      experiment = response['content']
      if experiment['experiment_report']:
        # download
        # # to html
        # # read report and transform to resource
        # if os.path.exists(os.path.join(target_path, experiment_name)):
        #   fp = open(os.path.join(target_path, experiment_name), 'rb')
        #   report_data = fp.read()
        #   report_data = loads(report_data)
        #   fp.close()
        #   # clear temp report file
        #   os.remove(os.path.join(target_path, experiment_name))
        #
        #   if len(report_data) == 0:
        #     print('no experiment reports')
        #     return
        #   for stage, stage_report in report_data.items():
        #     target_path = os.path.join(os.curdir, 'experiment', experiment_name, 'report', stage)
        #     if not os.path.exists(target_path):
        #       os.makedirs(target_path)
        #
        #     everything_to_html(stage_report, target_path)
        pass

      if experiment['experiment_model']:
        # download experiment model
        pass

  def process_dataset_command(self):
    dataset_name = FLAGS.dataset_name()
    if dataset_name is None:
      response = mlogger.getEnv().dashboard.dataset.get()
      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      datasets = response['content']
      table = PrettyTable(["name", "time", "task", "url", "publisher"])
      for dataset in datasets:
          table.add_row([
            dataset['dataset_name'],
            time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(dataset['dataset_time'])),
            dataset['task_num'],
            dataset['dataset_url'],
            dataset['dataset_publisher']
          ])

      print(table)
    else:
      response = mlogger.getEnv().dashboard.dataset.get(dataset_name=dataset_name)
      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      dataset = response['content']
      table = PrettyTable(["name", "time", "task", "url", "publisher"])
      table.add_row([
        dataset['dataset_name'],
        time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(dataset['dataset_time'])),
        dataset['task_num'],
        dataset['dataset_url'],
        dataset['dataset_publisher']
      ])
      print(table)

  def process_apply_command(self):
    task_name = FLAGS.task_name()
    apply_name = FLAGS.apply_name()
    if task_name is None:
      # list all applied task
      response = mlogger.getEnv().dashboard.apply.get()
      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      content = response['content']
      table = PrettyTable(["task", "apply", "time", "dataset", "experiment", "token"])

      for task in content:
        table.add_row([
          task['task_name'],
          task['apply_name'],
          time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(task['task_time'])),
          task['task_dataset'],
          task['task_experiment'],
          task['task_token']
        ])
      print(table)
    else:
      # apply a task
      if apply_name is None:
        apply_name = task_name

      response = mlogger.getEnv().dashboard.apply.post(apply_name=apply_name,
                               task_name=task_name)
      if response['status'] == "ERROR":
        logger.error(response['message'])
        return

      content = response['content']
      table = PrettyTable(["task", "apply", "time", "token"])
      table.add_row([
        content['task_name'],
        content['apply_name'],
        time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(content['apply_time'])),
        content['apply_token']
      ])
      print(table)

  def process_del_command(self):
    apply_name = FLAGS.apply_name()
    task_name = FLAGS.task_name()
    dataset_name = FLAGS.dataset_name()
    uuid = FLAGS.experiment_uuid()

    if apply_name is not None:
      response = mlogger.getEnv().dashboard.apply.delete(apply_name=apply_name)
      if response['status'] == 'ERROR':
        logger.error(response['message'])
        return

      logger.info(response['message'])

    if dataset_name is not None:
      response = mlogger.getEnv().dashboard.dataset.delete(dataset_name=dataset_name)
      if response['status'] == 'ERROR':
        logger.error(response['message'])
        return

      logger.info(response['message'])


    if task_name is not None:
      response = mlogger.getEnv().dashboard.task.delete(task_name=task_name)
      if response['status'] == 'ERROR':
        logger.error(response['message'])
        return

      logger.info(response['message'])

    if uuid is not None:
      response = mlogger.getEnv().dashboard.experiment.delete(experiment_uuid=uuid)
      if response['status'] == 'ERROR':
        logger.error(response['message'])
        return

      logger.info(response['message'])

  def _key_params(self):
    # related parameters
    # 0.step token
    token = FLAGS.token()
    if not PY3 and token is not None:
      token = unicode(token)
    token = self.app_token if token is None else token

    # 1.step check name, if None, set it as current time automatically
    name = FLAGS.name()
    if name is None:
      name = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime(time.time()))

    if not PY3:
      name = unicode(name)

    # 2.step check main folder (all related model code, includes main_file and main_param)
    main_folder = FLAGS.main_folder()
    if main_folder is None:
      main_folder = os.path.abspath(os.curdir)

    main_file = FLAGS.main_file()
    if main_file is None or not os.path.exists(os.path.join(main_folder, main_file)):
      logger.error('main executing file dont exist')
      return

    # 3.step check dump dir (all running data is stored here)
    dump_dir = FLAGS.dump()
    if dump_dir is None:
      dump_dir = os.path.join(os.path.abspath(os.curdir), 'dump')
      if not os.path.exists(dump_dir):
        os.makedirs(dump_dir)

    # 4.step what is task
    task = FLAGS.task()

    # 5.step model params
    main_param = FLAGS.main_param()

    return token, name, main_file, main_folder, dump_dir, main_param, task

  def process_challenge_command(self):
    token, name, main_file, main_folder, dump_dir, params, task = self._key_params()
    challenge_cmd = 'antgo challenge'
    if token is not None:
      challenge_cmd += ' --token=%s'%token

    if name is not None:
      challenge_cmd += ' --name=%s'%name

    if main_file is not None:
      challenge_cmd += ' --main_file=%s'%main_file

    if main_folder is not None:
      challenge_cmd += ' --main_folder=%s'%main_folder

    if params is not None:
      challenge_cmd += ' --main_param=%s'%params

    if task is not None:
      challenge_cmd += ' --task=%s' % task

    if dump_dir is not None:
      challenge_cmd += ' --dump=%s'%dump_dir

    challenge_cmd += ' > %s-challenge.log 2>&1 &' % name

    # launch background process
    subprocess.call("nohup %s"%challenge_cmd, shell=True)

  def process_train_command(self):
    token, name, main_file, main_folder, dump_dir, params, task = self._key_params()
    train_cmd = 'antgo train'
    if token is not None:
      train_cmd += ' --token=%s' % token

    if name is not None:
      train_cmd += ' --name=%s' % name

    if main_file is not None:
      train_cmd += ' --main_file=%s' % main_file

    if main_folder is not None:
      train_cmd += ' --main_folder=%s' % main_folder

    if params is not None:
      train_cmd += ' --main_param=%s' % params

    if task is not None:
      train_cmd += ' --task=%s' % task

    if dump_dir is not None:
      train_cmd += ' --dump=%s' % dump_dir

    train_cmd += ' > %s-train.log 2>&1 &' % name

    # launch background process
    subprocess.call("nohup %s" % train_cmd, shell=True)

  def process_cmd(self, command):
    try:
      if command == 'task':
        self.process_task_command()
      elif command == 'experiment':
        self.process_experiment_command()
      elif command == 'dataset':
        self.process_dataset_command()
      elif command == 'apply':
        self.process_apply_command()
      elif command == 'del':
        self.process_del_command()
      elif command == 'challenge':
        self.process_challenge_command()
      elif command == 'train':
        self.process_train_command()
    except:
      logger.error('error response from server')

  def start(self):
    cmd = ''
    if PY3:
      cmd = input('antgo > ')
    else:
      cmd = raw_input('antgo > ')

    while cmd != 'quit':
      try:
        command = cmd.split(' ')
        assert(command[0] in ['task',
                              'dataset',
                              'experiment',
                              'apply',
                              'del',
                              'challenge',
                              'train'])

        flags.cli_param_flags(command[1:])

        # process user command
        self.process_cmd(command[0])

        # clear flags
        flags.clear_cli_param_flags()
      except:
        logger.error('error antgo command\n')

      if PY3:
        cmd = input('antgo > ')
      else:
        cmd = raw_input('antgo > ')


# -*- coding: UTF-8 -*-
# @Time : 2018/8/24
# @File : coco2017.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import copy
import itertools
import json
import time
import numpy as np
from collections import defaultdict
import sys

PYTHON_VERSION = sys.version_info[0]
if PYTHON_VERSION == 2:
  from urllib import urlretrieve
elif PYTHON_VERSION == 3:
  from urllib.request import urlretrieve
from antgo.dataflow.dataset.dataset import *
from antgo.utils import mask as maskUtils
from antgo.utils.fs import download
from antgo.utils.fs import maybe_here_match_format
from antgo.utils import logger

__all__ = ['COCO2017']


class CocoAPI():
  def __init__(self, annotation_file=None):
    """
    Constructor of Microsoft COCO helper class for reading and visualizing annotations.
    :param annotation_file (str): location of annotation file
    :param image_folder (str): location to the folder that hosts images.
    :return:
    """
    # load dataset
    self.dataset, self.anns, self.cats, self.imgs = dict(), dict(), dict(), dict()
    self.imgToAnns, self.catToImgs = defaultdict(list), defaultdict(list)
    if not annotation_file == None:
      print('loading annotations into memory...')
      tic = time.time()
      dataset = json.load(open(annotation_file, 'r'))
      assert type(dataset) == dict, 'annotation file format {} not supported'.format(type(dataset))
      print('Done (t={:0.2f}s)'.format(time.time() - tic))
      self.dataset = dataset
      self.createIndex()

  def createIndex(self):
    # create index
    print('creating index...')
    anns, cats, imgs = {}, {}, {}
    imgToAnns, catToImgs = defaultdict(list), defaultdict(list)
    if 'annotations' in self.dataset:
      for ann in self.dataset['annotations']:
        if 'segments_info' not in ann:
          imgToAnns[ann['image_id']].append(ann)
          anns[ann['id']] = ann
        else:
          imgToAnns[ann['image_id']].append(ann['segments_info'])
          for mm in ann['segments_info']:
            anns[mm['id']] = mm

    if 'images' in self.dataset:
      for img in self.dataset['images']:
        imgs[img['id']] = img

    if 'categories' in self.dataset:
      for cat in self.dataset['categories']:
        cats[cat['id']] = cat

    if 'annotations' in self.dataset and 'categories' in self.dataset:
      for ann in self.dataset['annotations']:
        if 'segments_info' not in ann:
          catToImgs[ann['category_id']].append(ann['image_id'])
        else:
          for mm in ann['segments_info']:
            catToImgs[mm['category_id']] = ann['image_id']

    print('index created!')

    # create class members
    self.anns = anns
    self.imgToAnns = imgToAnns
    self.catToImgs = catToImgs
    self.imgs = imgs
    self.cats = cats

  def info(self):
    """
    Print information about the annotation file.
    :return:
    """
    for key, value in self.dataset['info'].items():
      print('{}: {}'.format(key, value))

  def getAnnIds(self, imgIds=[], catIds=[], areaRng=[], iscrowd=None):
    """
    Get ann ids that satisfy given filter conditions. default skips that filter
    :param imgIds  (int array)     : get anns for given imgs
           catIds  (int array)     : get anns for given cats
           areaRng (float array)   : get anns for given area range (e.g. [0 inf])
           iscrowd (boolean)       : get anns for given crowd label (False or True)
    :return: ids (int array)       : integer array of ann ids
    """
    imgIds = imgIds if type(imgIds) == list else [imgIds]
    catIds = catIds if type(catIds) == list else [catIds]

    if len(imgIds) == len(catIds) == len(areaRng) == 0:
      anns = self.dataset['annotations']
    else:
      if not len(imgIds) == 0:
        lists = [self.imgToAnns[imgId] for imgId in imgIds if imgId in self.imgToAnns]
        anns = list(itertools.chain.from_iterable(lists))
      else:
        anns = self.dataset['annotations']
      anns = anns if len(catIds) == 0 else [ann for ann in anns if ann['category_id'] in catIds]
      anns = anns if len(areaRng) == 0 else [ann for ann in anns if
                                             ann['area'] > areaRng[0] and ann['area'] < areaRng[1]]
    if not iscrowd == None:
      ids = [ann['id'] for ann in anns if ann['iscrowd'] == iscrowd]
    else:
      ids = [ann['id'] for ann in anns]
    return ids

  def getCatIds(self, catNms=[], supNms=[], catIds=[]):
    """
    filtering parameters. default skips that filter.
    :param catNms (str array)  : get cats for given cat names
    :param supNms (str array)  : get cats for given supercategory names
    :param catIds (int array)  : get cats for given cat ids
    :return: ids (int array)   : integer array of cat ids
    """
    catNms = catNms if type(catNms) == list else [catNms]
    supNms = supNms if type(supNms) == list else [supNms]
    catIds = catIds if type(catIds) == list else [catIds]

    if len(catNms) == len(supNms) == len(catIds) == 0:
      cats = self.dataset['categories']
    else:
      cats = self.dataset['categories']
      cats = cats if len(catNms) == 0 else [cat for cat in cats if cat['name'] in catNms]
      cats = cats if len(supNms) == 0 else [cat for cat in cats if cat['supercategory'] in supNms]
      cats = cats if len(catIds) == 0 else [cat for cat in cats if cat['id'] in catIds]
    ids = [cat['id'] for cat in cats]
    return ids

  def getImgIds(self, imgIds=[], catIds=[]):
    '''
    Get img ids that satisfy given filter conditions.
    :param imgIds (int array) : get imgs for given ids
    :param catIds (int array) : get imgs with all given cats
    :return: ids (int array)  : integer array of img ids
    '''
    imgIds = imgIds if type(imgIds) == list else [imgIds]
    catIds = catIds if type(catIds) == list else [catIds]

    if len(imgIds) == len(catIds) == 0:
      ids = self.imgs.keys()
    else:
      ids = set(imgIds)
      for i, catId in enumerate(catIds):
        if i == 0 and len(ids) == 0:
          ids = set(self.catToImgs[catId])
        else:
          ids &= set(self.catToImgs[catId])
    return list(ids)

  def loadAnns(self, ids=[]):
    """
    Load anns with the specified ids.
    :param ids (int array)       : integer ids specifying anns
    :return: anns (object array) : loaded ann objects
    """
    if type(ids) == list:
      return [self.anns[id] for id in ids]
    elif type(ids) == int:
      return [self.anns[ids]]

  def loadCats(self, ids=[]):
    """
    Load cats with the specified ids.
    :param ids (int array)       : integer ids specifying cats
    :return: cats (object array) : loaded cat objects
    """
    if type(ids) == list:
      return [self.cats[id] for id in ids]
    elif type(ids) == int:
      return [self.cats[ids]]

  def loadImgs(self, ids=[]):
    """
    Load anns with the specified ids.
    :param ids (int array)       : integer ids specifying img
    :return: imgs (object array) : loaded img objects
    """
    if type(ids) == list:
      return [self.imgs[id] for id in ids]
    elif type(ids) == int:
      return [self.imgs[ids]]

  def loadRes(self, resFile):
    """
    Load result file and return a result api object.
    :param   resFile (str)     : file name of result file
    :return: res (obj)         : result api object
    """
    res = CocoAPI()
    res.dataset['images'] = [img for img in self.dataset['images']]

    print('Loading and preparing results...')
    tic = time.time()
    if type(resFile) == str or type(resFile) == unicode:
      anns = json.load(open(resFile))
    elif type(resFile) == np.ndarray:
      anns = self.loadNumpyAnnotations(resFile)
    else:
      anns = resFile
    assert type(anns) == list, 'results in not an array of objects'
    annsImgIds = [ann['image_id'] for ann in anns]
    assert set(annsImgIds) == (set(annsImgIds) & set(self.getImgIds())), \
      'Results do not correspond to current coco set'
    if 'caption' in anns[0]:
      imgIds = set([img['id'] for img in res.dataset['images']]) & set([ann['image_id'] for ann in anns])
      res.dataset['images'] = [img for img in res.dataset['images'] if img['id'] in imgIds]
      for id, ann in enumerate(anns):
        ann['id'] = id + 1
    elif 'bbox' in anns[0] and not anns[0]['bbox'] == []:
      res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
      for id, ann in enumerate(anns):
        bb = ann['bbox']
        x1, x2, y1, y2 = [bb[0], bb[0] + bb[2], bb[1], bb[1] + bb[3]]
        if not 'segmentation' in ann:
          ann['segmentation'] = [[x1, y1, x1, y2, x2, y2, x2, y1]]
        ann['area'] = bb[2] * bb[3]
        ann['id'] = id + 1
        ann['iscrowd'] = 0
    elif 'segmentation' in anns[0]:
      res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
      for id, ann in enumerate(anns):
        # now only support compressed RLE format as segmentation results
        ann['area'] = maskUtils.area(ann['segmentation'])
        if not 'bbox' in ann:
          ann['bbox'] = maskUtils.toBbox(ann['segmentation'])
        ann['id'] = id + 1
        ann['iscrowd'] = 0
    elif 'keypoints' in anns[0]:
      res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
      for id, ann in enumerate(anns):
        s = ann['keypoints']
        x = s[0::3]
        y = s[1::3]
        x0, x1, y0, y1 = np.min(x), np.max(x), np.min(y), np.max(y)
        ann['area'] = (x1 - x0) * (y1 - y0)
        ann['id'] = id + 1
        ann['bbox'] = [x0, y0, x1 - x0, y1 - y0]
    print('DONE (t={:0.2f}s)'.format(time.time() - tic))

    res.dataset['annotations'] = anns
    res.createIndex()
    return res

  def download(self, tarDir=None, imgIds=[]):
    '''
    Download COCO images from mscoco.org server.
    :param tarDir (str): COCO results directory name
           imgIds (list): images to be downloaded
    :return:
    '''
    if tarDir is None:
      print('Please specify target directory')
      return -1
    if len(imgIds) == 0:
      imgs = self.imgs.values()
    else:
      imgs = self.loadImgs(imgIds)
    N = len(imgs)
    if not os.path.exists(tarDir):
      os.makedirs(tarDir)
    for i, img in enumerate(imgs):
      tic = time.time()
      fname = os.path.join(tarDir, img['file_name'])
      if not os.path.exists(fname):
        urlretrieve(img['coco_url'], fname)
      print('downloaded {}/{} images (t={:0.1f}s)'.format(i, N, time.time() - tic))

  def loadNumpyAnnotations(self, data):
    """
    Convert result data from a numpy array [Nx7] where each row contains {imageID,x1,y1,w,h,score,class}
    :param  data (numpy.ndarray)
    :return: annotations (python nested list)
    """
    print('Converting ndarray to lists...')
    assert (type(data) == np.ndarray)
    print(data.shape)
    assert (data.shape[1] == 7)
    N = data.shape[0]
    ann = []
    for i in range(N):
      if i % 1000000 == 0:
        print('{}/{}'.format(i, N))
      ann += [{
        'image_id': int(data[i, 0]),
        'bbox': [data[i, 1], data[i, 2], data[i, 3], data[i, 4]],
        'score': data[i, 5],
        'category_id': int(data[i, 6]),
      }]
    return ann

  def annToRLE(self, ann):
    """
    Convert annotation which can be polygons, uncompressed RLE to RLE.
    :return: binary mask (numpy 2D array)
    """
    t = self.imgs[ann['image_id']]
    h, w = t['height'], t['width']
    segm = ann['segmentation']
    if type(segm) == list:
      # polygon -- a single object might consist of multiple parts
      # we merge all parts into one mask rle code
      rles = maskUtils.frPyObjects(segm, h, w)
      rle = maskUtils.merge(rles)
    elif type(segm['counts']) == list:
      # uncompressed RLE
      rle = maskUtils.frPyObjects(segm, h, w)
    else:
      # rle
      rle = ann['segmentation']
    return rle

  def annToMask(self, ann):
    """
    Convert annotation which can be polygons, uncompressed RLE, or RLE to binary mask.
    :return: binary mask (numpy 2D array)
    """
    rle = self.annToRLE(ann)
    m = maskUtils.decode(rle)
    return m


class COCO2017(Dataset):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    '''
    'id': 0, 'supercategory': 'person', 'name': 'person'
    'id': 1, 'supercategory': 'vehicle', 'name': 'bicycle'
    'id': 2, 'supercategory': 'vehicle', 'name': 'car'
    'id': 3, 'supercategory': 'vehicle', 'name': 'motorcycle'
    'id': 4, 'supercategory': 'vehicle', 'name': 'airplane'
    'id': 5, 'supercategory': 'vehicle', 'name': 'bus'
    'id': 6, 'supercategory': 'vehicle', 'name': 'train'
    'id': 7, 'supercategory': 'vehicle', 'name': 'truck'
    'id': 8, 'supercategory': 'vehicle', 'name': 'boat'
    'id': 9, 'supercategory': 'outdoor', 'name': 'traffic light'
    'id': 10, 'supercategory': 'outdoor', 'name': 'fire hydrant'
    'id': 12, 'supercategory': 'outdoor', 'name': 'stop sign'
    'id': 13, 'supercategory': 'outdoor', 'name': 'parking meter'
    'id': 14, 'supercategory': 'outdoor', 'name': 'bench'
    'id': 15, 'supercategory': 'animal', 'name': 'bird'
    'id': 16, 'supercategory': 'animal', 'name': 'cat'
    'id': 17, 'supercategory': 'animal', 'name': 'dog'
    'id': 18, 'supercategory': 'animal', 'name': 'horse'
    'id': 19, 'supercategory': 'animal', 'name': 'sheep'
    'id': 20, 'supercategory': 'animal', 'name': 'cow'
    'id': 21, 'supercategory': 'animal', 'name': 'elephant'
    'id': 22, 'supercategory': 'animal', 'name': 'bear'
    'id': 23, 'supercategory': 'animal', 'name': 'zebra'
    'id': 24, 'supercategory': 'animal', 'name': 'giraffe'
    'id': 26, 'supercategory': 'accessory', 'name': 'backpack'
    'id': 27, 'supercategory': 'accessory', 'name': 'umbrella'
    'id': 30, 'supercategory': 'accessory', 'name': 'handbag'
    'id': 31, 'supercategory': 'accessory', 'name': 'tie'
    'id': 32, 'supercategory': 'accessory', 'name': 'suitcase'
    'id': 33, 'supercategory': 'sports', 'name': 'frisbee'
    'id': 34, 'supercategory': 'sports', 'name': 'skis'
    'id': 35, 'supercategory': 'sports', 'name': 'snowboard'
    'id': 36, 'supercategory': 'sports', 'name': 'sports ball'
    'id': 37, 'supercategory': 'sports', 'name': 'kite'
    'id': 38, 'supercategory': 'sports', 'name': 'baseball bat'
    'id': 39, 'supercategory': 'sports', 'name': 'baseball glove'
    'id': 40, 'supercategory': 'sports', 'name': 'skateboard'
    'id': 41, 'supercategory': 'sports', 'name': 'surfboard'
    'id': 42, 'supercategory': 'sports', 'name': 'tennis racket'
    'id': 43, 'supercategory': 'kitchen', 'name': 'bottle'
    'id': 45, 'supercategory': 'kitchen', 'name': 'wine glass'
    'id': 46, 'supercategory': 'kitchen', 'name': 'cup'
    'id': 47, 'supercategory': 'kitchen', 'name': 'fork'
    'id': 48, 'supercategory': 'kitchen', 'name': 'knife'
    'id': 49, 'supercategory': 'kitchen', 'name': 'spoon'
    'id': 50, 'supercategory': 'kitchen', 'name': 'bowl'
    'id': 51, 'supercategory': 'food', 'name': 'banana'
    'id': 52, 'supercategory': 'food', 'name': 'apple'
    'id': 53, 'supercategory': 'food', 'name': 'sandwich'
    'id': 54, 'supercategory': 'food', 'name': 'orange'
    'id': 55, 'supercategory': 'food', 'name': 'broccoli'
    'id': 56, 'supercategory': 'food', 'name': 'carrot'
    'id': 57, 'supercategory': 'food', 'name': 'hot dog'
    'id': 58, 'supercategory': 'food', 'name': 'pizza'
    'id': 59, 'supercategory': 'food', 'name': 'donut'
    'id': 60, 'supercategory': 'food', 'name': 'cake'
    'id': 61, 'supercategory': 'furniture', 'name': 'chair'
    'id': 62, 'supercategory': 'furniture', 'name': 'couch'
    'id': 63, 'supercategory': 'furniture', 'name': 'potted plant'
    'id': 64, 'supercategory': 'furniture', 'name': 'bed'
    'id': 66, 'supercategory': 'furniture', 'name': 'dining table'
    'id': 69, 'supercategory': 'furniture', 'name': 'toilet'
    'id': 71, 'supercategory': 'electronic', 'name': 'tv'
    'id': 72, 'supercategory': 'electronic', 'name': 'laptop'
    'id': 73, 'supercategory': 'electronic', 'name': 'mouse'
    'id': 74, 'supercategory': 'electronic', 'name': 'remote'
    'id': 75, 'supercategory': 'electronic', 'name': 'keyboard'
    'id': 76, 'supercategory': 'electronic', 'name': 'cell phone'
    'id': 77, 'supercategory': 'appliance', 'name': 'microwave'
    'id': 78, 'supercategory': 'appliance', 'name': 'oven'
    'id': 79, 'supercategory': 'appliance', 'name': 'toaster'
    'id': 80, 'supercategory': 'appliance', 'name': 'sink'
    'id': 81, 'supercategory': 'appliance', 'name': 'refrigerator'
    'id': 83, 'supercategory': 'indoor', 'name': 'book'
    'id': 84, 'supercategory': 'indoor', 'name': 'clock'
    'id': 85, 'supercategory': 'indoor', 'name': 'vase'
    'id': 86, 'supercategory': 'indoor', 'name': 'scissors'
    'id': 87, 'supercategory': 'indoor', 'name': 'teddy bear'
    'id': 88, 'supercategory': 'indoor', 'name': 'hair drier'
    'id': 89, 'supercategory': 'indoor', 'name': 'toothbrush'

    :param train_or_test: train, test or val
    :param dir:
    :param ext_params: year, annotation_type, included, excluded, transform

    '''
    super(COCO2017, self).__init__(train_or_test, dir, ext_params)
    self.year = "2017"
    self.dir = dir
    self.train_or_test = train_or_test
    self.data_type = None
    self.is_support_random_split = False
    self.task_type = getattr(self, 'task_type', None)
    self.task_type_subset = getattr(self, 'task_type_subset', 'stuff')
    self.task_test = getattr(self, 'task_test', None)

    assert(self.train_or_test in ['sample', 'train', 'val', 'test'])
    assert (self.task_type in ['SEGMENTATION', 'OBJECT-DETECTION', 'INSTANCE-SEGMENTATION', 'LANDMARK'])

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self._image_index = self.load_samples()
      return

    data_type = None
    if self.train_or_test == "train":
      data_type = "train" + self.year
    elif self.train_or_test == "val":
      data_type = "val" + self.year
    elif self.train_or_test == "test":
      data_type = "test" + self.year
    self.data_type = data_type

    if self.train_or_test in ['train', 'val']:
      if self.task_type in ["OBJECT-DETECTION", "INSTANCE-SEGMENTATION"]:
        # annotation file
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        # parse (for object detector)
        self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
        self.cat_ids = self.coco_api.getCatIds(self.cats)
        self.img_ids = self.coco_api.getImgIds(self.cat_ids)
      elif self.task_type == "LANDMARK":
        # parse (for object detector)
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        # parse (for person task)
        self.cat_ids = self.coco_api.getCatIds(catNms=['person'])
        self.img_ids = self.coco_api.getImgIds(catIds=self.cat_ids)

        # annotation file
        kps_ann_file = self.config_ann_file(data_type, self.dir, 'PersonKeypoint')
        self.coco_kps_api = CocoAPI(kps_ann_file)
      elif self.task_type == "IMAGE_CAPTION":
        # parse (for object detector)
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
        self.cat_ids = self.coco_api.getCatIds(self.cats)
        self.img_ids = self.coco_api.getImgIds(self.cat_ids)

        # parse (for caption)
        caps_ann_file = self.config_ann_file(data_type, self.dir, 'Caption')
        self.coco_caps_api = CocoAPI(caps_ann_file)
      elif self.task_type == "SEGMENTATION":
        # Panoptic Segmentation
        assert(self.task_type_subset in ['stuff', 'panoptic'])
        if self.task_type_subset == 'stuff':
          ann_file = os.path.join(self.dir, 'annotations', 'stuff_%s2017.json'%self.train_or_test)
          self.coco_api = CocoAPI(ann_file)

          self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
          self.cat_ids = self.coco_api.getCatIds(self.cats)
          self.img_ids = self.coco_api.getImgIds(self.cat_ids)
        else:
          # ann_file = os.path.join(self.dir, 'annotations', 'panoptic_%s2017.json'%self.train_or_test)
          # self.coco_api = CocoAPI(ann_file)
          # self.img_ids = self.coco_api.getImgIds()
          logger.error('not support')
          exit(-1)
      else:
        # annotation file
        ann_file = self.config_ann_file(data_type, self.dir, "Instance")
        self.coco_api = CocoAPI(ann_file)

        # parse (for object detector)
        self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
        self.cat_ids = self.coco_api.getCatIds(self.cats)
        self.img_ids = self.coco_api.getImgIds(self.cat_ids)
    else:
      ann_file = ''

      if self.task_test == 'dev':
        ann_file = os.path.join(self.dir, 'annotations', 'image_info_test-dev2017.json')
      else:
        ann_file = os.path.join(self.dir, 'annotations', 'image_info_test2017.json')

      self.coco_api = CocoAPI(ann_file)

      # parse (for object detector)
      self.cats = self.coco_api.loadCats(self.coco_api.getCatIds())
      self.cat_ids = self.coco_api.getCatIds(self.cats)
      self.img_ids = self.coco_api.getImgIds(self.cat_ids)

  def config_ann_file(self, data_type, data_dir, annotation_type):
    maybe_data_dir = maybe_here_match_format(data_dir, 'annotations')
    assert maybe_data_dir is not None

    ann_file = None
    if annotation_type == "Instance":
      ann_file = '%s/annotations/instances_%s.json' % (maybe_data_dir, data_type)
    elif annotation_type == "PersonKeypoint":
      ann_file = '%s/annotations/person_keypoints_%s.json' % (maybe_data_dir, data_type)
    elif annotation_type == "Caption":
      ann_file = '%s/annotations/captions_%s.json' % (maybe_data_dir, data_type)

    return ann_file

  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.copy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      idxs = copy.deepcopy(self.img_ids)
      if self.rng:
        self.rng.shuffle(idxs)

      for img_id in idxs:
        img_obj = self.coco_api.loadImgs(img_id)[0]
        if self.train_or_test == 'test':
          img = imread(os.path.join(self.dir, '%s2017' % self.train_or_test, img_obj['file_name']))
          yield [img, {}]
          continue

        if self.task_type == 'OBJECT-DETECTION' or \
                self.task_type=='INSTANCE-SEGMENTATION':
          annotation_ids = self.coco_api.getAnnIds(imgIds=img_obj['id'])
          annotation = self.coco_api.loadAnns(annotation_ids)
          img_annotation = {}

          num_objs = len(annotation)
          boxes = np.zeros((num_objs, 4), dtype=np.uint16)
          category_id = np.zeros((num_objs), dtype=np.int32)
          category = []
          supercategory = []
          area = np.zeros((num_objs), dtype=np.float32)
          segmentation = []

          category_id_name = {cc['id']: cc['name'] for cc in self.coco_api.dataset['categories']}
          category_id_supername = {cc['id']: cc['supercategory'] for cc in self.coco_api.dataset['categories']}

          for ix, obj in enumerate(annotation):
            x, y, w, h = obj['bbox']
            boxes[ix, :] = [x, y, x + w, y + h]

            category_id[ix] = obj['category_id']
            category.append(category_id_name[obj['category_id']])
            supercategory.append(category_id_supername[obj['category_id']])
            area[ix] = obj['area']
            segmentation.append(self.coco_api.annToMask(obj))

          img_annotation['bbox'] = boxes
          img_annotation['category_id'] = category_id
          img_annotation['category'] = category
          img_annotation['supercategory'] = supercategory
          img_annotation['flipped'] = False
          img_annotation['area'] = area
          img_annotation['segmentation'] = segmentation

          img = imread(os.path.join(self.dir, '%s2017'%self.train_or_test, img_obj['file_name']))
          segmentation_map = np.zeros((img.shape[0], img.shape[1]), dtype=np.int32)
          for ix, obj_seg in enumerate(segmentation):
            obj_id = category_id[ix]
            segmentation_map[np.where(obj_seg == 1)] = obj_id

          img_annotation.update({'segmentation_map': segmentation_map, 'file_id': img_obj['file_name']})
          img_annotation['info'] = (img.shape[0], img.shape[1], img.shape[2])
          yield [img, img_annotation]
        elif self.task_type == 'IMAGE_CAPTION':
          img = imread(os.path.join(self.dir, '%s2017'%self.train_or_test, img_obj['file_name']))
          ann_ids = self.coco_caps_api.getAnnIds(imgIds=img_obj['id'])
          anns = self.coco_caps_api.loadAnns(ann_ids)
          img_annotation = {'caption': [cap['caption'] for cap in anns], 'file_id': img_obj['file_name']}
          yield [img, img_annotation]
        elif self.task_type == 'SEGMENTATION':
          # stuff, Panoptic
          assert(self.task_type_subset in ['stuff', 'panoptic'])
          if self.task_type_subset == 'stuff':
            img = imread(os.path.join(self.dir, '%s2017'%self.train_or_test, img_obj['file_name']))
            annotation_ids = self.coco_api.getAnnIds(imgIds=img_obj['id'])
            annotation = self.coco_api.loadAnns(annotation_ids)
            img_annotation = {}
            supercategory = []
            num_objs = len(annotation)
            category_id = np.zeros((num_objs), dtype=np.int32)
            category = []
            category_id_name = {cc['id']: cc['name'] for cc in self.coco_api.dataset['categories']}
            category_id_supername = {cc['id']: cc['supercategory'] for cc in self.coco_api.dataset['categories']}

            segmentation = []
            for ix, obj in enumerate(annotation):
              category_id[ix] = obj['category_id']
              category.append(category_id_name[obj['category_id']])
              supercategory.append(category_id_supername[obj['category_id']])
              segmentation.append(self.coco_api.annToMask(obj))

            segmentation_map = np.zeros((img.shape[0], img.shape[1]), dtype=np.int32)
            for ix, obj_seg in enumerate(segmentation):
              obj_id = category_id[ix]
              segmentation_map[np.where(obj_seg == 1)] = obj_id

            img_annotation['category_id'] = category_id
            img_annotation['category'] = category
            img_annotation['supercategory'] = supercategory
            img_annotation['flipped'] = False
            img_annotation['segmentation'] = segmentation
            img_annotation['segmentation_map'] = segmentation_map
            img_annotation['file_id'] = img_obj['file_name']

            yield [img, img_annotation]
          else:
            yield [None, None]
        elif self.task_type == 'LANDMARK':
          img = imread(os.path.join(self.dir, '%s2017'%self.train_or_test, img_obj['file_name']))
          ann_ids = self.coco_kps_api.getAnnIds(imgIds=img_obj['id'])
          anns = self.coco_kps_api.loadAnns(ann_ids)

          img_annotation = {'keypoints': [], 'segmentation': []}
          person_bboxs = np.zeros((len(anns), 4))
          area = np.zeros((len(anns)))
          for person_index, person_ann in enumerate(anns):
            keypionts = person_ann['keypoints']
            segmentation = person_ann['segmentation']
            area[person_index] = person_ann['area']

            xx = keypionts[0::3]
            yy = keypionts[1::3]
            visible = keypionts[2::3]
            obj_kk = np.zeros((len(xx),3))
            obj_kk[:, 0] = xx
            obj_kk[:, 1] = yy
            obj_kk[:, 2] = visible
            img_annotation['keypoints'].append(obj_kk)
            img_annotation['segmentation'].append(segmentation)
            person_bboxs[person_index, :] = person_ann['bbox']

          img_annotation['bbox'] = person_bboxs
          img_annotation['area'] = area
          img_annotation['flipped'] = False
          img_annotation['file_id'] = img_obj['file_name']

          yield [img, img_annotation]
        else:
          img = imread(os.path.join(self.dir, '%s2017' % self.train_or_test, img_obj['file_name']))
          yield [img, None]

  @property
  def size(self):
    return len(self.img_ids)

  def split(self, train_validation_ratio=0.0, is_stratified_sampling=True):
    assert(self.train_or_test == 'train')
    validation_coco = COCO2017('val', self.dir, self.ext_params)
    return self, validation_coco


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: gpu.py
# Author: jian(jian@mltalker.com)
from __future__ import unicode_literals

import os
import re
from antgo.utils.utils import change_env
import subprocess

__all__ = ['change_gpu', 'get_nr_gpu', 'get_gpus','gpu_running_info']


def change_gpu(val):
  return change_env('CUDA_VISIBLE_DEVICES', str(val))


def get_nr_gpu():
    env = os.environ.get('CUDA_VISIBLE_DEVICES', None)
    if env is None:
        return 0

    return len(env.split(','))


def get_gpus():
    """ return a list of GPU physical id"""
    env = os.environ.get('CUDA_VISIBLE_DEVICES', None)
    if env is None:
        return ''
    return map(int, env.strip().split(','))


def gpu_running_info(pid=None):
    try:
        content = subprocess.check_output('nvidia-smi').decode('utf-8')
    except:
        return None

    driver_version = re.findall('(?<=Driver Version: )[\d.]+', content)[0]
    gpu_basic_info = re.findall('(?<=\|)[ ]+\d+%[ ]+\d+C[]+\w+\d+W / /d+W[ ]+(?=\|)', content)
    is_plan_b = False
    if len(gpu_basic_info) == 0:
        gpu_basic_info = re.findall('(?<=\|)[ ]+N/A[ ]+\d+C[]+\w+\d+W / /d+W[ ]+(?=\|)', content)
        is_plan_b = True
    
    gpu_num = len(gpu_basic_info)
    gpus=[]

    for gpu_index in range(gpu_num):
        result = re.split('\s+', gpu_basic_info[gpu_index].strip())
        gpus.append(result[2])

    gpu_pwr_info = re.findall('\d+W / \d+W',content)
    gpu_pwr_usage=[]
    gpu_pwr_cap=[]
    for gpu_index in range(gpu_num):
        result = re.split('/',gpu_pwr_info[gpu_index])
        pwr_usage = re.findall('\d+',result[0])[0]
        pwr_cap = re.findall('\d+',result[1])[0]
        gpu_pwr_usage.append(int(pwr_usage))
        gpu_pwr_cap.append(int(pwr_cap))

    gpu_mem_info = re.findall('\d+MiB / \d+MiB',content)
    gpu_mem_usage=[]
    gpu_mem_max=[]
    for gpu_index in range(gpu_num):
        result = re.split('/',gpu_mem_info[gpu_index])
        mem_usage = re.findall('\d+',result[0])[0]
        mem_max = re.findall('\d+',result[1])[0]
        gpu_mem_usage.append(int(mem_usage))
        gpu_mem_max.append(int(mem_max))

    gpu_util = re.findall('\d+(?=%)',content)
    gpu_util = [int(util) for util in gpu_util]
    if not is_plan_b:
        gpu_util = [int(util) for id, util in enumerate(gpu_util) if id % 2 == 1]

    # occupy_gpus = []
    # if pid is not None:
    #     pattern = '\d\s+(?={pid})'.format(pid=pid)
    #     terms = re.findall(pattern,content)
    #     for term in terms:
    #         occupy_gpus.append(int(term))

    occupy_gpus = list(range(len(gpus)))

    return {'gpus': gpus,
            'driver-version': driver_version,
            'gpu_pwr_usage': gpu_pwr_usage,
            'gpu_pwr_cap': gpu_pwr_cap,
            'gpu_mem_usage': gpu_mem_usage,
            'gpu_mem_max': gpu_mem_max,
            'gpu_util': gpu_util,
            'occupy_gpus': occupy_gpus}


# encoding=utf-8
# @Time    : 17-8-10
# @File    : config.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
try:
  import xml.etree.cElementTree as ET
except ImportError:
  import xml.etree.ElementTree as ET
import os

class _Config(object):
  def __init__(self):
    self._attribs = {}

  def parse_xml(self, config_xml):
    tree = ET.ElementTree(file=config_xml)
    root = tree.getroot()

    for child in root:
      if child.tag == 'subgradientserver':
        subgradientserver_config = {}
        for subgradientserver_child in child:
          val = subgradientserver_child.text.strip() if subgradientserver_child.text is not None else None
          if subgradientserver_child.tag == 'subgradientserver_port':
            val = int(val)
          subgradientserver_config[subgradientserver_child.tag] = val

        setattr(self, child.tag, subgradientserver_config)
        continue

      val = child.text.strip() if child.text is not None else None
      setattr(self, child.tag, val)

      if child.tag == 'factory' and val is not None and val != '':
        setattr(self, 'data_factory', os.path.join(val, 'dataset'))
        setattr(self, 'task_factory', os.path.join(val, 'task'))

      self._attribs[child.tag] = val

  def write_xml(self, config_xml, attribs={}):
    all_attribs = attribs
    all_attribs.update(self._attribs)

    tree = ET.ElementTree(file=config_xml)
    root = tree.getroot()
    for child in root:
      if child.tag in all_attribs:
        child.text = all_attribs[child.tag]

    tree.write(config_xml, encoding="utf-8",xml_declaration=True)


AntConfig = _Config()


# -*- coding: UTF-8 -*-
# @Time    : 17-12-27
# @File    : celeba.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset import *
import os
import numpy as np
__all__ = ['CelebA']

CELEBA_URL = "http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html"
class CelebA(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(CelebA, self).__init__(train_or_test, dir, ext_params=params)

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    # 0.step maybe download
    if not (os.path.exists(os.path.join(self.dir, 'Img')) and
              os.path.exists(os.path.join(self.dir, 'Anno')) and
              os.path.exists(os.path.join(self.dir, 'Eval'))):
      self.download(self.dir,
                    file_names=['pack.zip'],
                    default_url='shell:unzip {file_placeholder} -d %s'%'/'.join(self.dir.split('/')[0:-1]))
    
    # 1.step data folder (wild or align)
    image_flag = getattr(self, 'image', 'align')
    if image_flag == 'align':
      self._data_folder = os.path.join(self.dir, 'Img', 'img_align_celeba')
    else:
      self._data_folder = os.path.join(self.dir, 'Img', 'img_celeba')
      
    # 2.step data split(train/val/test)
    self._file_list = []
    self._file_map = {}
    with open(os.path.join(self.dir, 'Eval', 'list_eval_partition.txt'), 'r') as fp:
      content = fp.readline()
      while content:
        file_name, file_flag = content.split(' ')
        if self.train_or_test == 'train':
          if int(file_flag) == 0:
            self._file_map[file_name] = len(self._file_list)
            self._file_list.append(os.path.join(self._data_folder, file_name))
        elif self.train_or_test == 'val':
          if int(file_flag) == 1:
            self._file_map[file_name] = len(self._file_list)
            self._file_list.append(os.path.join(self._data_folder, file_name))
        else:
          if int(file_flag) == 2:
            self._file_map[file_name] = len(self._file_list)
            self._file_list.append(os.path.join(self._data_folder, file_name))
        
        content = fp.readline()
    
    self.ids = list(range(len(self._file_list)))
    
    # 3.step annotation folder
    self._annotations = {}
    # 3.1 step landmark annotation
    # landmark name
    # lefteye_x
    # lefteye_y
    # righteye_x
    # righteye_y
    # nose_x
    # nose_y
    # leftmouth_x
    # leftmouth_y
    # rightmouth_x
    # rightmouth_y
    if image_flag == 'align':
      with open(os.path.join(self.dir, 'Anno', 'list_landmarks_align_celeba.txt')) as fp:
        # skip fist two rows
        fp.readline()
        fp.readline()
        landmark_content = fp.readline()
        while landmark_content:
          file_name, l_x, l_y, r_x, r_y, n_x, n_y, lm_x, lm_y, rm_x, rm_y = \
            [i for i in landmark_content.replace('\n','').split(' ') if i != '']

          if file_name in self._file_map:
            self._annotations[file_name] = {'landmark': [int(l_x),
                                                         int(l_y),
                                                         int(r_x),
                                                         int(r_y),
                                                         int(n_x),
                                                         int(n_y),
                                                         int(lm_x),
                                                         int(lm_y),
                                                         int(rm_x),
                                                         int(rm_y)]}
          landmark_content = fp.readline()
    else:
      with open(os.path.join(self.dir, 'Anno', 'list_landmarks_celeba.txt')) as fp:
        # skip fist two rows
        fp.readline()
        fp.readline()
        landmark_content = fp.readline()
        while landmark_content:
          file_name, l_x, l_y, r_x, r_y, n_x, n_y, lm_x, lm_y, rm_x, rm_y = \
            [i for i in landmark_content.replace('\n','').split(' ') if i != '']
          if file_name in self._file_map:
            self._annotations[file_name] = {'landmark': [int(l_x),
                                                         int(l_y),
                                                         int(r_x),
                                                         int(r_y),
                                                         int(n_x),
                                                         int(n_y),
                                                         int(lm_x),
                                                         int(lm_y),
                                                         int(rm_x),
                                                         int(rm_y)]}
          landmark_content = fp.readline()

    # 3.2 step attribute annotation
    # attribute names
    # 5_o_Clock_Shadow
    # Arched_Eyebrows
    # Attractive
    # Bags_Under_Eyes
    # Bald
    # Bangs
    # Big_Lips
    # Big_Nose
    # Black_Hair
    # Blond_Hair
    # Blurry
    # Brown_Hair
    # Bushy_Eyebrows
    # Chubby
    # Double_Chin
    # Eyeglasses
    # Goatee
    # Gray_Hair
    # Heavy_Makeup
    # High_Cheekbones
    # Male
    # Mouth_Slightly_Open
    # Mustache
    # Narrow_Eyes
    # No_Beard
    # Oval_Face
    # Pale_Skin
    # Pointy_Nose
    # Receding_Hairline
    # Rosy_Cheeks
    # Sideburns
    # Smiling
    # Straight_Hair
    # Wavy_Hair
    # Wearing_Earrings
    # Wearing_Hat
    # Wearing_Lipstick
    # Wearing_Necklace
    # Wearing_Necktie
    # Young

    self.attribute_name_dict = {'5_o_Clock_Shadow': 0,
                                'Arched_Eyebrows': 1,
                                'Attractive': 2,
                                'Bags_Under_Eyes': 3,
                                'Bald': 4,
                                'Bangs': 5,
                                'Big_Lips': 6,
                                'Big_Nose': 7,
                                'Black_Hair': 8,
                                'Blond_Hair': 9,
                                'Blurry': 10,
                                'Brown_Hair': 11,
                                'Bushy_Eyebrows': 12,
                                'Chubby': 13,
                                'Double_Chin': 14,
                                'Eyeglasses': 15,
                                'Goatee': 16,
                                'Gray_Hair': 17,
                                'Heavy_Makeup': 18,
                                'High_Cheekbones': 19,
                                'Male': 20,
                                'Mouth_Slightly_Open': 21,
                                'Mustache': 22,
                                'Narrow_Eyes': 23,
                                'No_Beard': 24,
                                'Oval_Face': 25,
                                'Pale_Skin': 26,
                                'Pointy_Nose': 27,
                                'Receding_Hairline': 28,
                                'Rosy_Cheeks': 29,
                                'Sideburns': 30,
                                'Smiling': 31,
                                'Straight_Hair': 32,
                                'Wavy_Hair': 33,
                                'Wearing_Earrings': 34,
                                'Wearing_Hat': 35,
                                'Wearing_Lipstick': 36,
                                'Wearing_Necklace': 37,
                                'Wearing_Necktie': 38,
                                'Young': 39}

    with open(os.path.join(self.dir, 'Anno', 'list_attr_celeba.txt')) as fp:
      # skip first two rows
      fp.readline()
      fp.readline()

      attribute_content = fp.readline()
      while attribute_content:
        file_name, \
        o_Clock_Shadow, \
        Arched_Eyebrows, \
        Attractive, \
        Bags_Under_Eyes, \
        Bald, \
        Bangs, \
        Big_Lips, \
        Big_Nose, \
        Black_Hair, \
        Blond_Hair, \
        Blurry, \
        Brown_Hair, \
        Bushy_Eyebrows, \
        Chubby, \
        Double_Chin, \
        Eyeglasses, \
        Goatee, \
        Gray_Hair, \
        Heavy_Makeup, \
        High_Cheekbones, \
        Male, \
        Mouth_Slightly_Open, \
        Mustache, \
        Narrow_Eyes, \
        No_Beard, \
        Oval_Face, \
        Pale_Skin, \
        Pointy_Nose, \
        Receding_Hairline, \
        Rosy_Cheeks, \
        Sideburns, \
        Smiling, \
        Straight_Hair, \
        Wavy_Hair, \
        Wearing_Earrings, \
        Wearing_Hat, \
        Wearing_Lipstick, \
        Wearing_Necklace, \
        Wearing_Necktie, \
        Young = [i for i in attribute_content.replace('\n', '').split(' ') if i != '']
        
        if file_name in self._file_map:
          self._annotations[file_name]['attribute_name'] = self.attribute_name_dict
          self._annotations[file_name]['attribute'] = [int(o_Clock_Shadow),
                                                       int(Arched_Eyebrows),
                                                       int(Attractive),
                                                       int(Bags_Under_Eyes),
                                                       int(Bald),
                                                       int(Bangs),
                                                       int(Big_Lips),
                                                       int(Big_Nose),
                                                       int(Black_Hair),
                                                       int(Blond_Hair),
                                                       int(Blurry),
                                                       int(Brown_Hair),
                                                       int(Bushy_Eyebrows),
                                                       int(Chubby),
                                                       int(Double_Chin),
                                                       int(Eyeglasses),
                                                       int(Goatee),
                                                       int(Gray_Hair),
                                                       int(Heavy_Makeup),
                                                       int(High_Cheekbones),
                                                       int(Male),
                                                       int(Mouth_Slightly_Open),
                                                       int(Mustache),
                                                       int(Narrow_Eyes),
                                                       int(No_Beard),
                                                       int(Oval_Face),
                                                       int(Pale_Skin),
                                                       int(Pointy_Nose),
                                                       int(Receding_Hairline),
                                                       int(Rosy_Cheeks),
                                                       int(Sideburns),
                                                       int(Smiling),
                                                       int(Straight_Hair),
                                                       int(Wavy_Hair),
                                                       int(Wearing_Earrings),
                                                       int(Wearing_Hat),
                                                       int(Wearing_Lipstick),
                                                       int(Wearing_Necklace),
                                                       int(Wearing_Necktie),
                                                       int(Young) ]
          
        attribute_content = fp.readline()
    # 3.3 step bounding box
    # bounding box: x_1 y_1 width height
    with open(os.path.join(self.dir, 'Anno', 'list_bbox_celeba.txt')) as fp:
      # skip first two rows
      fp.readline()
      fp.readline()
      
      bounding_box_content = fp.readline()
      while bounding_box_content:
        file_name, x, y, width, height = [i for i in bounding_box_content.replace('\n', '').split(' ') if i != '']
        
        if file_name in self._file_map:
          boxes = np.zeros((1, 4), dtype=np.uint16)
          boxes[0, 0] = int(x)
          boxes[0, 1] = int(y)
          boxes[0, 2] = int(x) + int(width)
          boxes[0, 3] = int(y) + int(height)
          self._annotations[file_name]['bbox'] = boxes
        bounding_box_content = fp.readline()

    # 3.4 step person label
    with open(os.path.join(self.dir, 'Anno', 'identity_CelebA.txt')) as fp:
      content = fp.readline()
      while content:
        content = content.strip()
        a, b = content.split(' ')
        if a in self._annotations:
          self._annotations[a]['category_id'] = int(b)
          self._annotations[a]['category'] = b

        content = fp.readline()

  @property
  def size(self):
    return len(self.ids)
  
  def data_pool(self):
    if self.train_or_test == 'sample':
      # self.sample_data_pool(self.data_samples, self.ids)
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)

      for k in idxs:
        image_file = self._file_list[k]
        file_name = image_file.split('/')[-1]
        image = imread(image_file)

        self._annotations[file_name].update({'id': k, 'info': [image.shape[0], image.shape[1], image.shape[2]]})
        self._annotations[file_name].update({'name': file_name})
        yield [image, self._annotations[file_name]]

  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    image_file = self._file_list[id]
    file_name = image_file.split('/')[-1]
    image = imread(image_file)
    self._annotations[file_name].update({'id': id, 'info': [image.shape[0], image.shape[1], image.shape[2]]})
    self._annotations[file_name].update({'name': file_name})
    return image, self._annotations[file_name]
  
  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    validation_dataset = CelebA('val', self.dir)
    return self, validation_dataset

  def attribute_name_map(self, name):
    if name not in self.attribute_name_dict:
      return -1
    else:
      return self.attribute_name_dict[name]



# -*- coding: UTF-8 -*-
# @Time    : 2021/1/8 7:23 下午
# @File    : jupytercontext.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.context import *
from antgo.ant.train import *
from antgo.ant.challenge import *
from antgo.ant.demo import *
from antgo.ant.browser import *
from antgo.ant.batch import *


class JupyterContext(Context):
  def __init__(self):
    super(JupyterContext, self).__init__()

  def __prepare_header(self, api_token, name, param, **kwargs):
    # 1.step 获取基本配置信息
    config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
    Config.parse_xml(config_xml)

    factory = getattr(Config, 'factory', None)
    assert factory is not None

    if not os.path.exists(Config.data_factory):
      os.makedirs(Config.data_factory)
    if not os.path.exists(Config.task_factory):
      os.makedirs(Config.task_factory)

    self.name = name
    self.data_factory = Config.data_factory
    self.task_factory = Config.task_factory
    self.main_folder = os.path.abspath(os.curdir)

    if type(param) == str:
      param = yaml.load(open(param, 'r'))
      self.params = param
    else:
      self.params = param

    # 2.step 设置来自实验名称
    dump_dir = os.path.join(os.path.abspath(os.curdir), 'dump')
    if not os.path.exists(dump_dir):
      os.makedirs(dump_dir)

    from_experiment = kwargs.get('from_experiment', None)
    if from_experiment is not None:
      experiment_path = os.path.join(dump_dir, from_experiment)
      if not os.path.exists(experiment_path):
        logger.error('couldnt find experiment %s' % from_experiment)
        raise IOError

      if os.path.exists(os.path.join(experiment_path, 'train')):
        self.from_experiment = os.path.join(experiment_path, 'train')
      elif os.path.exists(os.path.join(experiment_path, 'inference')):
        self.from_experiment = os.path.join(experiment_path, 'inference')
      else:
        self.from_experiment = experiment_path

    return dump_dir

  @contextmanager
  def Train(self, api_token, exp_name, exp_param, **kwargs):
    # 1.step 配置基本信息
    dump_dir = self.__prepare_header(api_token, exp_name, exp_param, **kwargs)
    # 2.step 创建时间戳
    time_stamp = timestamp()
    # 3.step 获得是否制定自定义任务
    task = kwargs.get('task', None)
    # 4.step 是否跳过训练过程
    skip_training = kwargs.get('skip_training', False)
    # 5.step 设置设备
    devices = kwargs.get('devices', '')

    # 6.step 创建训练过程调度对象
    train = AntTrain(self,
                     exp_name,
                     self.data_factory,
                     dump_dir,
                     api_token,
                     task,
                     skip_training=skip_training,
                     time_stamp=time_stamp,
                     devices=devices,
                     main_folder=self.main_folder)

    yield train

    # 日志更新
    mlogger.update()


  @contextmanager
  def Challenge(self, api_token, exp_name, exp_param, benchmark='', **kwargs):
    # 1.step 配置基本信息
    dump_dir = self.__prepare_header(api_token, exp_name, exp_param, **kwargs)
    # 2.step 创建时间戳
    time_stamp = timestamp()
    # 3.step 获得是否制定自定义任务
    task = kwargs.get('task', None)
    # 4.step 设置设备
    devices = kwargs.get('devices', '')

    # 5.step 创建挑战过程调度对象
    challenge = AntChallenge(self,
                             exp_name,
                             self.data_factory,
                             dump_dir,
                             api_token,
                             task,
                             benchmark,
                             time_stamp=time_stamp,
                             devices=devices,
                             main_folder=self.main_folder)
    yield challenge

    # 日志更新
    mlogger.update()

  @contextmanager
  def Demo(self, api_token, exp_name, exp_param, html_template=None, ip=None, port=None, **kwargs):
    # 1.step 配置基本信息
    dump_dir = self.__prepare_header(api_token, exp_name, exp_param, **kwargs)
    # 2.step 创建时间戳
    time_stamp = timestamp()
    # 3.step 获得是否制定自定义任务
    task = kwargs.get('task', None)
    # 4.step 设置设备
    devices = kwargs.get('devices', '')

    demo = AntDemo(self,
                   exp_name,
                   dump_dir,
                   api_token,
                   task,
                   html_template=html_template,
                   ip=ip,
                   port=port,
                   time_stamp=time_stamp,
                   devices=devices)
    yield demo

    # 日志更新
    mlogger.update()


  @contextmanager
  def Browser(self, api_token, exp_name, exp_param, ip=None, port=10000, **kwargs):
    # 1.step 配置基本信息
    dump_dir = self.__prepare_header(api_token, exp_name, exp_param, **kwargs)

    browser = AntBrowser(self,
                         exp_name,
                         api_token,
                         ip,
                         port,
                         self.data_factory,
                         kwargs.get('dataset', ''),
                         dump_dir)

    yield browser

    # 日志更新
    mlogger.update()


  @contextmanager
  def Activelearning(self):
    yield

  @contextmanager
  def Generate(self):
    yield

  @contextmanager
  def Batch(self, api_token, exp_name, exp_param, ip=None, port=10000, unlabel=False, **kwargs):
    # 1.step 配置基本信息
    dump_dir = self.__prepare_header(api_token, exp_name, exp_param, **kwargs)
    # 2.step 自定义任务
    task = kwargs.get('task', None)

    batch = AntBatch(self,
                     exp_name,
                     ip,
                     port,
                     api_token,
                     self.data_factory,
                     dump_dir,
                     task,
                     unlabel=unlabel,
                     restore_experiment=kwargs.get('restore_experiment', None))
    yield batch

    # 日志更新
    mlogger.update()



# -*- coding: UTF-8 -*-
# @Time    : 17-11-27
# @File    : sandbox.py
# @Author  : Jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os, sys
from contextlib import contextmanager
from antgo.utils.concurrency import *
from antgo.utils import logger
import antvis.client.mlogger as mlogger
import multiprocessing
import time


@contextmanager
def running_sandbox(*wargs, **kwargs):
  def _sandbox_thread(**kwargs):
    if 'sandbox_time' in kwargs and kwargs['sandbox_time'] is not None:
      launch_time = kwargs['sandbox_launch_time']
      sandbox_running_time = kwargs['sandbox_time']
      # now time
      now_time = time.time()
      if now_time - launch_time > sandbox_running_time:
        # arriving custom max running time
        logger.info('max running time is arriving, exit antgo running environment')

        # experiment result save process
        if 'sandbox_dump_dir' in kwargs and \
                'sandbox_experiment' in kwargs and \
                'sandbox_user_token' in kwargs:
          # TODO 当实验运行超时时，执行固定操作
          pass

        # exit globally
        mlogger.exit()
        os._exit(0)

  try:
    # timer thread
    timer_thread = None

    sandbox_thread_kwargs = kwargs
    sandbox_thread_kwargs['sandbox_launch_time'] = time.time()
    if 'sandbox_time' in kwargs and kwargs['sandbox_time'] is not None:
      sandbox_time = kwargs['sandbox_time']
      # check time format (s,m,h,d)
      time_unit = sandbox_time[-1]
      if time_unit not in ['s', 'm', 'h', 'd']:
        logger.error('sand time isnt correct')
        exit(-1)
      time_value = sandbox_time[:-1]
      time_value = int(time_value)

      second = time_value
      if time_unit == 'm':
        second = second * 60
      elif time_unit == 'h':
        second = second * 3600
      elif time_unit == 'd':
        second = second * 3600 * 24
      running_time = second

      sandbox_thread_kwargs['sandbox_time'] = running_time

    if 'sandbox_time' in sandbox_thread_kwargs:
      timer_thread = TimerThread([lambda: _sandbox_thread(**sandbox_thread_kwargs)], periodic=60)
      # start thread
      timer_thread.start()

    yield

    # 线程退出
    if timer_thread is not None:
      # stop thread
      timer_thread.stop()
      timer_thread.join()

    logger.info('exit antgo running environment')
  except:
    logger.error('error occured in antgo running environment')


# -*- coding: UTF-8 -*-
# @Time    : 2019-04-17 22:21
# @File    : person_search_task.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.measures.base import *
from antgo.utils._bbox import bbox_overlaps
from antgo.task.task import *


default={'AntPersonSearchCMC': ('CMC', 'PERSON_SEARCH')}


class AntPersonSearchCMC(AntMeasure):
  def __init__(self, task):
    super(AntPersonSearchCMC, self).__init__(task, "CMC")
    assert(task.task_type == 'PERSON_SEARCH')

    self.support_rank_index = 0
    self.is_support_rank = True
    self.top_k = 1

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    bingo_num = 0
    pool_size = 0
    for det_predict, det_gt in data:
      # detected bbox
      det_bbox = det_predict['det-bbox']
      # l2 distance between det bbox and gt bbox
      det_score = np.array(det_predict['det-score'])
      det_sorted_inds = np.argsort(det_score, axis=1).tolist()

      gt_bbox = det_gt['bbox']
      gt_category = det_gt['category_id']
      overlaps = bbox_overlaps(
        np.ascontiguousarray(det_bbox, dtype=np.float),
        np.ascontiguousarray(gt_bbox, dtype=np.float))

      pool_size += len(gt_bbox)
      gtm = np.ones((len(gt_bbox))) * (-1)
      for dind, d in enumerate(det_bbox):
        best_match = -1
        iou_thres = 0.5
        for gind, g in enumerate(gt_bbox):
          if gtm[gind] >= 0:
            continue
          if overlaps[dind, gind] < iou_thres:
            continue

          iou_thres = overlaps[dind, gind]
          best_match = gind

        if best_match > -1:
          gtm[best_match] = dind
          if gt_category[best_match] in det_sorted_inds[dind][0:self.top_k]:
            bingo_num += 1

    cmc = float(bingo_num)/float(pool_size)
    return {'statistic': {'name': self.name,
                          'value': [{'name': 'CMC',
                                     'value': cmc,
                                     'type': 'SCALAR',
                                     'x': '',
                                     'y': ''},
                                   ]}}


if __name__ == '__main__':
  num = 20
  data = []
  for _ in range(num):
    x1 = np.random.random()
    y1 = np.random.random()

    x2 = x1 + np.random.random()
    y2 = y1 + np.random.random()

    det_predict = {'det-bbox': [[x1, y1, x2, y2]],
                   'det-score': np.random.random((1, 10)).tolist()}

    det_gt = {
      'bbox': [[x1,y1,x2,y2]]*10,
      'category_id':[3]*10
    }

    data.append((det_predict, det_gt))

  cmc = AntPersonSearchCMC(None)
  result = cmc.eva(data,None)
  pass


# encoding=utf-8
# @Time    : 17-5-15
# @File    : moving_statistic.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np


class MovingStatistic(object):
  def __init__(self):
    self.value = None

  def add(self, value):
    pass

  def get(self):
    pass


class MovingAverage(MovingStatistic):
  def __init__(self, ws=1):
    super(MovingAverage, self).__init__()
    self.value = np.zeros((ws))
    self.window_size = ws
    self.count = 0
    
  def add(self, value):
    index = self.count % self.window_size
    self.value[index] = value
    
    self.count = index + 1

  def get(self):
    if len(self.value) == 0:
      return None
    return np.mean(self.value)


#  -*- coding: UTF-8 -*-
#  File: __init__.py
#  Author: jian(jian@mltalker.com)
from __future__ import unicode_literals

from pkgutil import walk_packages
import os

def _global_import(name):
    p = __import__(name, globals(), None, level=1)
    lst = p.__all__ if '__all__' in dir(p) else dir(p)
    del globals()[name]
    for k in lst:
        globals()[k] = p.__dict__[k]

_global_import('utils')


# -*- coding: UTF-8 -*-
# Time: 12/31/17
# File: dht.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import time
import requests
import subprocess
import json
import traceback
import threading
import sys
from antgo.dataflow.basic import *
from antgo.utils.serialize import *
import numpy as np
import shutil
from antgo.ant.warehouse import *

from antgo import config
PYTHON_VERSION = sys.version_info[0]
if PYTHON_VERSION == 2:
    import Queue as queue
elif PYTHON_VERSION == 3:
    import queue as queue

Config = config.AntConfig


class FuncThread(threading.Thread):
  def __init__(self, func, blocks, data_q):
    threading.Thread.__init__(self)
    self._func = func
    self._blocks = blocks
    self._data_q = data_q
    self.daemon = True

  def run(self):
    self._func(self._blocks, self._data_q)


def dataset_upload_dht(dataset_name, data_generators, dump_dir):
  dataset_readme = {}

  if not os.path.exists(dump_dir):
    os.makedirs(dump_dir)

  # parse dataset and split to blocks
  for dataset_c, dataset_content in data_generators.items():
    # data sample generator
    dataset_generator = dataset_content['generator']
    # data sample number
    dataset_num = dataset_content['num']
    # block number
    block_num = dataset_content['block']

    # rectify block number
    if block_num > dataset_num:
      block_num = 1

    dataset_index = np.arange(0, int(dataset_num)).tolist()

    # prepare block content
    blocks_content = []
    blocks_name = ['%s_%s_block_%d'%(dataset_name, dataset_c, bi) for bi in range(block_num)]
    blocks_index = {}

    block_size = int(int(dataset_num) // int(block_num))

    current_block_index = 0
    for index, content in zip(dataset_index, dataset_generator):
      data, label = content
      block_index = int(index // block_size)
      if block_index == block_num:
        block_index = block_num - 1

      if current_block_index != block_index:
        with open(os.path.join(dump_dir, blocks_name[current_block_index]), 'wb') as fp:
          fp.write(dumps(blocks_content))

        blocks_content = []
        current_block_index = block_index

      # block content
      blocks_content.append((index, data, label))

      # sample index
      if blocks_name[block_index] not in blocks_index:
        blocks_index[blocks_name[block_index]] = []

      blocks_index[blocks_name[block_index]].append(index)

    with open(os.path.join(dump_dir, blocks_name[current_block_index]), 'wb') as fp:
      fp.write(dumps(blocks_content))

    # dataset sample number
    dataset_readme['%s_count'%dataset_c] = dataset_num
    # dataset blocks
    dataset_readme['%s_block'%dataset_c] = blocks_name
    # dataset block sample index
    dataset_readme['%s_index'%dataset_c] = blocks_index

  # save readme
  with open(os.path.join(dump_dir, 'readme.yaml'), 'w') as fp:
    yaml.safe_dump(dataset_readme, fp)

  # upload to dht
  short_dir = dump_dir.split('/')[-1]
  # s = subprocess.Popen('ipfs add -r %s'%short_dir,
  #                      shell=True,
  #                      cwd='/'.join(dump_dir.split('/')[0:-1]), stdout=subprocess.PIPE)
  # result = s.communicate()[0].decode()
  #
  # dataset_hash_code = None
  # for ff in result.split('\n'):
  #   if len(ff) > 0:
  #     cmd, hash_code, file_h = ff.split(' ')
  #     if file_h == short_dir:
  #       dataset_hash_code = hash_code
  #       break
  import ipfsapi
  ipfs_host = os.environ.get('IPFS_HOST', '127.0.0.1')
  ipfs = ipfsapi.connect(ipfs_host, 5001)

  dataset_hash_code = ''
  result = ipfs.add(dump_dir, recursive=True)
  for ff in result:
    if ff['Name'] == short_dir:
      dataset_hash_code = ff['Hash']
      break

  return dataset_hash_code


def dataset_download_dht(dataset_folder, train_or_test, dht_address, data_queue, db, threads_num):
  # 1.step download dataset readme.yaml
  import ipfsapi
  ipfs_host = os.environ.get('IPFS_HOST', '127.0.0.1')
  ipfs = ipfsapi.connect(ipfs_host, 5001)

  if not os.path.exists(os.path.join(dataset_folder,'temp')):
    os.makedirs(os.path.join(dataset_folder,'temp'))

  os.chdir(os.path.join(dataset_folder,'temp'))
  ipfs.get('%s/readme.yaml' % dht_address)

  with open(os.path.join(dataset_folder, 'temp', 'readme.yaml'),'r') as fp:
    readme_config = yaml.load(fp)
    # notify dataset sample number
    data_queue.put({'count': readme_config['%s_count'%train_or_test]})

  # all data blocks and their index
  block_list = readme_config['%s_block'%train_or_test]
  block_index_list = readme_config['%s_index'%train_or_test]

  # 2.step download data block in parallel
  ok_block_list = []
  ok_index_list = []

  for b in block_list:
    if db.get(b) is not None:
      ok_block_list.append(b)
      ok_index_list.extend(block_index_list[b])

  # notify prepared block index
  if len(ok_index_list) > 0:
    data_queue.put(ok_index_list)

  # prepare waiting blocks
  block_address_list = [(block_index, '%s/%s' % (dht_address, block_name)) for block_index, block_name in
                        enumerate(block_list) if block_name not in ok_block_list]

  if len(block_address_list) == 0:
    return

  # continue download data from dht
  def _pull_from_dht(block_addresses, data_q):
    ipfs_host = os.environ.get('IPFS_HOST', '127.0.0.1')
    local_ipfs = ipfsapi.connect(ipfs_host, 5001)

    for block_index, block_address in block_addresses:
      try:
        # 1.step pull
        local_ipfs.get(block_address)

        # 3.step read content from block
        data_q.put((block_index, block_address.split('/')[-1]))
      except:
        pass

    data_q.put(None)

  waiting_block_num = len(block_address_list)
  if waiting_block_num < threads_num:
    threads_num = waiting_block_num

  blocks_in_thread = int(int(waiting_block_num) / int(threads_num))
  assign_responsability_blocks = [None for _ in range(threads_num)]
  for s in range(threads_num - 1):
    assign_responsability_blocks[s] = block_address_list[s*blocks_in_thread:(s+1)*blocks_in_thread]

  assign_responsability_blocks[threads_num-1] = block_address_list[(threads_num-1)*blocks_in_thread:]

  q = queue.Queue()
  # initialize thread
  process_threads = [FuncThread(_pull_from_dht, assign_responsability_blocks[i], q) for i in range(threads_num)]
  # launch thread
  for pt in process_threads:
    pt.start()

  none_num = threads_num
  while True:
    content = q.get()
    if content == None:
      none_num -= 1
      if none_num == 0:
        break

    if content is not None:
      block_index, block_file = content

      ok_samples = []
      with open(os.path.join(dataset_folder, 'temp', block_file), 'rb') as fp:
        # parse content
        binary_content = fp.read()
        data_list = loads(binary_content)

        for a,b,c in data_list:
          ok_samples.append((a, b, c, block_file))

      # notfiy prepared block
      if len(ok_samples) > 0:
        data_queue.put(ok_samples)

  # wating until all stop
  for pt in process_threads:
    pt.join()


def experiment_download_dht(dump_dir, experiment, token, proxy=None, signature='123',target='qiniu', address=None):
  # call in an independent process
  try:
    if experiment is None:
      kterms = dump_dir.split('/')
      dump_dir = '/'.join(kterms[:-1])
      experiment = kterms[-1]

    experiment_path = os.path.join(dump_dir, experiment)
    if not os.path.exists(experiment_path):
      os.makedirs(experiment_path)

    if token is not None:
      # 1.step get experiment address
      user_authorization = {'Authorization': "token " + token}

      server = getattr(Config, 'server_ip', 'www.mltalker.com')
      port = getattr(Config, 'server_port', '8999')
      res = requests.get('http://%s:%s/hub/api/experiment/%s/address' % (server, port, experiment),
                         headers=user_authorization)
      experiment_info = json.loads(res.content)

      address = experiment_info['ADDRESS']
      address_update_time = experiment_info['ADDRESS_UPDATE_TIME']
    elif proxy is not None:
      res = requests.get('http://%s/update/model/%s/' % (proxy, experiment),
                    data={'signature': signature})
      experiment_info = json.loads(res.content)
      if experiment_info['code'] != 'Success':
        logger.error('couldnt get experiment address')
        return False

      address = experiment_info['address']

    if address is None:
      logger.error('couldnt download experiment')
      return False

    # 2.step get from dht
    try:
      if target == 'ipfs':
        import ipfsapi
        os.chdir(experiment_path)
        ipfs_host = os.environ.get('IPFS_HOST', '127.0.0.1')
        ipfs = ipfsapi.connect(ipfs_host, 5001)
        ipfs.get(address)
      elif target == 'qiniu':
        os.chdir(experiment_path)
        result = qiniu_download(address, experiment_path)
        if result is None:
          return False
        address = result.split('/')[-1]
      else:
        logger.error('dont support target %s platform'%target)
        return False
    except:
      logger.error('couldnt get experiment address')
      return False

    # 3.step post process (rename and extract)
    # 3.1.step rename
    rename_shell = 'mv %s %s.tar.gz'%(address, experiment)
    subprocess.call(rename_shell, shell=True, cwd=experiment_path)
    # 3.2.step untar
    untar_shell = 'openssl enc -d -aes256 -in %s.tar.gz -k %s | tar xz -C %s'%(experiment, signature, '.')
    subprocess.call(untar_shell, shell=True, cwd=experiment_path)
    # 3.3.step clear tar file
    os.remove(os.path.join(experiment_path, '%s.tar.gz'%experiment))
    return True
  except:
    # remove unincomplete experiment
    shutil.rmtree(os.path.join(dump_dir, experiment))
    # error info
    traceback.print_exc()
    return False


def experiment_upload_dht(dump_dir, experiment, token, proxy=None, signature='123', bucket='experiment', target='qiniu'):
  # call in an independent process
  try:
    if token is None and proxy is None:
      return

    if experiment is None:
      kterms = dump_dir.split('/')
      dump_dir = '/'.join(kterms[:-1])
      experiment = kterms[-1]

    if not os.path.exists(os.path.join(dump_dir, experiment)):
      logger.error('experiment %s does not exist'%experiment)
      return

    # 1.step tar all files
    tar_shell = 'tar -czf - * | openssl enc -e -aes256 -out %s.tar.gz -k %s' % (experiment, signature)
    subprocess.call(tar_shell, shell=True, cwd=os.path.join(dump_dir, experiment))

    # 2.step ipfs add to dht
    experiment_hash = ''
    if target == 'ipfs':
      import ipfsapi
      ipfs_host = os.environ.get('IPFS_HOST', '127.0.0.1')
      ipfs = ipfsapi.connect(ipfs_host, 5001)
      res = ipfs.add(os.path.join(os.path.join(dump_dir, experiment), '%s.tar.gz' % experiment))
      experiment_hash = res['Hash']
    elif target == 'qiniu':
      result = qiniu_upload(os.path.join(dump_dir, experiment, '%s.tar.gz'%experiment), bucket=bucket, max_size=500)
      if result is None:
        logger.error('couldnt upload experiment %s to qiniu/%s'%(experiment, bucket))
        return

      experiment_hash = result
    else:
      # dont support now
      logger.error('dont support target %s platform'%target)
      return

    # 3.step notify mltalker
    try:
      if token is not None:
        user_authorization = {'Authorization': "token " + token}

        server = getattr(Config, 'server_ip', 'www.mltalker.com')
        port = getattr(Config, 'server_port', '8999')
        requests.patch('http://%s:%s/hub/api/experiment/%s/address' % (server, port, experiment),
                       data={'address': experiment_hash, 'time': time.time()},
                       headers=user_authorization)
      elif proxy is not None:
        requests.post('http://%s/update/model/%s/'%(proxy, experiment),
                      data={'signature': signature, 'address': experiment_hash})
    except:
      logger.error('couldnt notify server')

    # 4.step clear
    os.remove(os.path.join(dump_dir, experiment, '%s.tar.gz' % experiment))
    return True
  except:
    traceback.print_exc()
    return False


# test update experiment record
# experiment_upload_dht('/Users/jian/PycharmProjects/antgo/antgo/example/',
#                       'de03d0b6-9a2f-4704-94f0-a48cda92d8d5-20181219-143919-050970',
#                       token=None,
#                       proxy='127.0.0.1:10001',
#                       signature='aaa'
#                       )

# test download experiment record
# experiment_download_dht('/Users/Jian/Downloads/',
#                         'de03d0b6-9a2f-4704-94f0-a48cda92d8d5-20181219-143919-050970',
#                         token=None,
#                         proxy='127.0.0.1:10001',
#                         signature='aaa')


# encoding=utf-8
# @Time    : 17-6-13
# @File    : recorder.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import tarfile
import numpy as np
from six.moves import range
import xml.etree.ElementTree as ET

from antgo.utils import logger, get_rng, memoized
from antgo.utils.fs import mkdir_p, download, maybe_here
from antgo.utils.timer import timed_operation
from .dataset import *
__all__ = ['ILSVRC12']


CAFFE_ILSVRC12_URL = "http://dl.caffe.berkeleyvision.org/caffe_ilsvrc12.tar.gz"
class ILSVRCMeta(object):
  """
  Some metadata for ILSVRC dataset.
  """
  def __init__(self, dir=None):
    # todo: not completed
    self.dir = dir

    f = os.path.join(self.dir, 'synsets.txt')
    if not os.path.isfile(f):
        self._download_caffe_meta()

  def get_synset_words_1000(self):
    """
    :returns a dict of {cls_number: cls_name}
    """
    fname = os.path.join(self.dir, 'synset_words.txt')
    assert os.path.isfile(fname)
    lines = [x.strip() for x in open(fname).readlines()]
    return dict(enumerate(lines))

  def get_synset_1000(self):
    """
    :returns a dict of {cls_number: synset_id}
    """
    fname = os.path.join(self.dir, 'synsets.txt')
    assert os.path.isfile(fname)
    lines = [x.strip() for x in open(fname).readlines()]
    inv_synset = dict([ (b, a) for a, b in enumerate(lines)])
    return dict(enumerate(lines)), inv_synset

  def _download_caffe_meta(self):
    fpath = download(CAFFE_ILSVRC12_URL, self.dir)
    tarfile.open(fpath, 'r:gz').extractall(self.dir)


class ILSVRC12(Dataset):
  def __init__(self, train_or_test, dataset_path, params=None):
    """
    :param dataset_path: A directory containing a subdir named `name`, where the
        original ILSVRC12_`name`.tar gets decompressed.
    :param train_or_test: 'train' or 'val' or 'test'

    Dir should have the following structure:

    .. code-block:: none

        dir/
          train/
            n02134418/
              n02134418_198.JPEG
              ...
            ...
          val/
            ILSVRC2012_val_00000001.JPEG
            ...
          test/
            ILSVRC2012_test_00000001.JPEG
            ...
          bbox/
            n02134418/
              n02134418_198.xml
              ...
            ...

    After decompress ILSVRC12_img_train.tar, you can use the following
    command to build the above structure for `train/`:

    .. code-block:: none
        tar xvf ILSVRC12_img_train.tar -C train && cd train
        find -type f -name '*.tar' | parallel -P 10 'echo {} && mkdir -p {/.} && tar xf {} -C {/.}'
        Or:
        for i in *.tar; do dir=${i%.tar}; echo $dir; mkdir -p $dir; tar xf $i -C $dir; done

    """
    # init parent class
    super(ILSVRC12, self).__init__(train_or_test, dataset_path)
    self.train_or_test = train_or_test

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    # some meta info
    if not os.path.exists(os.path.join(self.dir, 'meta')):
      self.download(os.path.join(self.dir, 'meta'),
                    default_url="http://dl.caffe.berkeleyvision.org/caffe_ilsvrc12.tar.gz",
                    auto_untar=True,
                    is_gz=True)
    
    meta = ILSVRCMeta(os.path.join(self.dir, 'meta'))
    self.synset, self.inv_synset = meta.get_synset_1000()
    
    # image list
    bbdir = os.path.join(self.dir, 'bbox')
    if self.train_or_test == 'train':
      # maybe download
      if not os.path.exists(os.path.join(self.dir, self.train_or_test)):
        self.download(self.dir, file_names=['ILSVRC2012_img_train.tar'],
          default_url="shell:mkdir train && tar xvf {file_placeholder} -C train\n cd train && find -type f -name '*.tar' | parallel -P 10 'echo {} && mkdir -p {/.} && tar xf {} -C {/.}'")
  
      self.imglist = self.get_train_image_list(self.train_or_test)
      self.bblist, _ = ILSVRC12.get_bbox(bbdir, self.imglist, self.train_or_test)
    else:
      # maybe download
      if train_or_test == 'val':
        if not os.path.exists(os.path.join(self.dir, self.train_or_test)):
          self.download(self.dir, file_names=['ILSVRC2012_img_val.tar'],
            default_url="shell:mkdir train && tar xvf {file_placeholder} -C train\n cd train && find -type f -name '*.tar' | parallel -P 10 'echo {} && mkdir -p {/.} && tar xf {} -C {/.}'")
    
      self.imglist = self.get_val_or_test_image_list(self.train_or_test)
      if self.train_or_test == 'val':
        self.bblist, label_list = ILSVRC12.get_bbox(bbdir, self.imglist, self.train_or_test)
        self.imglist = [(self.imglist[k][0], self.inv_synset[label_list[k]]) for k in range(len(self.imglist))]

    self.ids = list(range(len(self.imglist)))

  def get_train_image_list(self, name):
    """
    :param name: 'train'
    :returns: list of (image filename, cls)
    """
    image_list = [('%s/%s'%(class_name, file_name), self.inv_synset[class_name]) \
                   for class_index, class_name in enumerate(os.listdir(os.path.join(self.dir, name)) ) \
                   if os.path.isdir(os.path.join(self.dir, name, class_name)) \
                   for file_name in os.listdir(os.path.join(self.dir, name, class_name))]

    return image_list
  def get_val_or_test_image_list(self, name):
    """
    :param name: 'val'
    :returns: list of (image filename, cls)
    """
    image_list = [(filename,-1) for fileindex,filename in enumerate(os.listdir(os.path.join(self.dir, self.train_or_test, name))) ]
    return image_list

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    validation_dataset = ILSVRC12('val', self.dir)
    return self, validation_dataset

  @property
  def size(self):
    return len(self.ids)

  def data_pool(self):
    """
    Produce original images of shape [h, w, 3], and label,
    and optionally a bbox of [xmin, ymin, xmax, ymax]
    """
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      idxs = np.arange(len(self.imglist))
      if self.rng:
          self.rng.shuffle(idxs)

      for k in idxs:
        fname, label = self.imglist[k]
        fname = os.path.join(self.dir, self.train_or_test, fname)
        im = imread(fname.strip())

        if im.ndim == 2:
          im = np.expand_dims(im, 2).repeat(3, 2)

        if self.train_or_test == "train" or \
                self.train_or_test == 'val':
          bbox = self.bblist[k]
          if bbox is None:
              bbox = np.array([0, 0, im.shape[1] - 1, im.shape[0] - 1])

          yield [im, {'category_id': label,
                      'bbox': bbox,
                      'id': k,
                      'info': (im.shape[0], im.shape[1], im.shape[2])}]
        else:
          yield [im]
  
  def at(self, k):
    if self.train_or_test == 'sample':
      return self.data_samples[k]

    fname, label = self.imglist[k]
    fname = os.path.join(self.dir, self.train_or_test, fname)
    im = imread(fname.strip())
  
    if im.ndim == 2:
      im = np.expand_dims(im, 2).repeat(3, 2)
  
    if self.train_or_test == "train" or \
            self.train_or_test == 'val':
      bbox = self.bblist[k]
      if bbox is None:
        bbox = np.array([0, 0, im.shape[1] - 1, im.shape[0] - 1])
    
      return [im, {'category_id': label,
                  'bbox': bbox,
                  'id': k,
                  'info': (im.shape[0], im.shape[1], im.shape[2])}]
    else:
      return [im]

  @staticmethod
  def get_bbox(bbox_dir, imglist, train_or_test):
    ret = []
    ret_label = []
    def parse_bbox(fname):
      root = ET.parse(fname).getroot()
      size = root.find('size').getchildren()
      size = map(int, [size[0].text, size[1].text])

      box = root.find('object').find('bndbox').getchildren()
      box = map(lambda x: float(x.text), box)

      label = root.find('object').find('name').text
      return np.asarray(list(box), dtype='float32'),label

    with timed_operation('Loading Bounding Boxes ...'):
      cnt = 0
      import tqdm
      for k in tqdm.trange(len(imglist)):
        fname = imglist[k][0]
        fname = fname[:-4] + 'xml'
        fname = os.path.join(bbox_dir,train_or_test ,fname)
        try:
          box,label = parse_bbox(fname)
          ret.append(box)
          ret_label.append(label)
          cnt += 1
        except KeyboardInterrupt:
          raise
        except:
          ret.append(None)
          ret_label.append(-1)
      logger.info("{}/{} images have bounding box.".format(cnt, len(imglist)))
    return ret, ret_label



# -*- coding: UTF-8 -*-
# @Time    : 2019/1/23 7:12 PM
# @File    : activelearning_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
from antgo.crowdsource.base_server import *
import tornado.web
import os
import shutil
import json
import numpy as np
import uuid
import signal
import functools
import base64
import time


class UploadHandler(BaseHandler):
  def post(self):
    # 1.step 获取第几轮索引
    round_index = self.db.get('round', -1)
    if round_index == -1:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'no round in db')
      return

    # 2.step 保存上传数据(已标注)
    slice_index = self.get_argument('sliceIndex')
    slice_index = int(slice_index)
    slice_num = self.get_argument('sliceNum')
    slice_num = int(slice_num)

    file_metas = self.request.files['file']
    filename = ''
    for meta in file_metas:
      filename = meta['filename']
      filepath = os.path.join(self.upload_folder, filename)
      if slice_index == 0:
        with open(filepath, 'wb') as fp:
          fp.write(meta['body'])
      else:
        with open(filepath, 'ab') as fp:
          fp.write(meta['body'])

      self.write('finished!')

    if slice_index == slice_num - 1:
      # 3.step 修改当前状态
      self.db["process_state"] = 'LABEL-CHECK'

      # 4.step notify backend
      self.request_queue.put({'ROUND': round_index, 'FILE': filename})


class DownloadHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    # 1.step 获取第几轮索引
    round_index = self.db.get('round', -1)
    if round_index == -1:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID, 'no round in db')
      return

    # 2.step 下载数据(未标注)
    unlabeled_dataset = self.db.get('unlabel_dataset', None)
    if unlabeled_dataset is None:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND, 'unlabeled data dont exist')
      return

    if not os.path.exists(os.path.join(self.download_folder, unlabeled_dataset)):
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND, 'unlabeled data dont exist')
      return

    # 3.step download
    # Content-Type
    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', 'attachment; filename=' + unlabeled_dataset)

    buffer_size = 64 * 1024
    with open(os.path.join(self.download_folder, unlabeled_dataset), 'rb') as fp:
      content = fp.read()
      content_size = len(content)
      buffer_segments = content_size // buffer_size
      for buffer_seg_i in range(buffer_segments):
        buffer_data = content[buffer_seg_i * buffer_size: (buffer_seg_i + 1) * buffer_size]
        yield self.write(buffer_data)

      yield self.write(content[buffer_segments * buffer_size:])

    # 4.step 返回
    self.finish()


class ActiveLearningState(BaseHandler):
  def patch(self, *args, **kwargs):
    # 1.step update round
    round_index = self.get_argument('round', None)
    if round_index is not None:
      self.db['round'] = int(round_index)

    # 2.step update processing state
    process_state = self.get_argument('process_state', None)
    if process_state is not None:
      self.db["process_state"] = process_state

      if process_state == 'LABEL-FINISH':
        next_round_waiting = self.get_argument('next_round_waiting', None)
        if next_round_waiting is None:
          self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
          return

        self.db["next_round_waiting"] = next_round_waiting
        self.db['time'] = time.time()

      if process_state == 'UNLABEL-RESET':
        unlabel_dataset = self.get_argument('unlabel_dataset', None)
        if unlabel_dataset is None:
          self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
          return

        self.db['unlabel_dataset'] = unlabel_dataset

    # 3.step unlabeled data number
    unlabeled_num = self.get_argument('unlabeled_size', None)
    if unlabeled_num is not None:
      self.db['unlabeled_size'] = unlabeled_num

    # 4.step labeled data number
    labeled_num = self.get_argument('labeled_size', None)
    if labeled_num is not None:
      self.db['labeled_size'] = labeled_num

    # 5.step round data number
    round_size = self.get_argument('round_size', None)
    if round_size is not None:
      self.db['round_size'] = round_size

    self.response(RESPONSE_STATUS_CODE.RESOURCE_SUCCESS_CREATED)


class IndexHandler(BaseHandler):
  def get(self):
    process_state = self.db.get('process_state', "UNLABEL-PREPARE")
    info = {'STATE': process_state}
    if process_state == 'UNLABEL-PREPARE' or process_state == "LABEL-FINISH":
      waiting_time = self.db.get('next_round_waiting', 0)
      start_time = self.db.get('time', 0)
      if waiting_time == 0 or start_time == 0:
        info['WAITING_TIME'] = 0
      else:
        info['WAITING_TIME'] = (waiting_time - (time.time() - start_time)) / 60

    if process_state == 'LABEL-ERROR':
      info['MESSAGE'] = "label error, please update label"

    labeled_size = self.get_argument('labeled_size', 0)
    unlabeled_size = self.get_argument('unlabeled_size', 0)
    round_size = self.get_argument('round_size', 0)
    info['LABELED_SIZE'] = labeled_size
    info['UNLABELED_SIZE'] = unlabeled_size
    info['ROUND_SIZE'] = round_size
    info['FINISHED_ROUND'] = self.db.get('round', 0)
    info['PERFORMANCE'] = 0

    info.update(self.keywords_template)

    return self.render(self.html_template, **info)


def activelearning_web_server(activelearning_name,
                              main_folder,
                              html_template,
                              keywords_template,
                              task,
                              server_port,
                              parent_id,
                              download_folder,
                              upload_folder,
                              request_queue):
  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  client_socket = None
  tornado.options.parse_command_line()

  if not os.path.exists(os.path.join(main_folder, 'web', 'template')):
    os.makedirs(os.path.join(main_folder, 'web', 'template'))

  activelearning_server_template_dir = os.path.join(main_folder, 'web', 'template')

  if not os.path.exists(os.path.join(main_folder, 'web', 'static')):
    os.makedirs(os.path.join(main_folder, 'web', 'static'))

  activelearning_server_static_dir = os.path.join(main_folder, 'web', 'static')

  static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
  for static_file in os.listdir(os.path.join(static_folder, 'resource', 'static')):
    if static_file[0] == '.':
      continue
    if not os.path.isfile(os.path.join(static_folder, 'resource', 'static', static_file)):
      continue

    shutil.copy(os.path.join(static_folder, 'resource', 'static', static_file),
                activelearning_server_static_dir)

  if html_template is None:
    shutil.copy(os.path.join(static_folder, 'resource', 'templates', 'activelearning.html'),
                activelearning_server_template_dir)
    html_template = 'activelearning.html'
  else:
    shutil.copy(os.path.join(main_folder, html_template), activelearning_server_template_dir)

  # data folder
  if not os.path.exists(os.path.join(main_folder, 'web', 'static', 'data')):
    os.makedirs(os.path.join(main_folder, 'web', 'static', 'data'))

  if not os.path.exists(os.path.join(main_folder, 'web', 'static', 'data', 'annotations')):
    os.makedirs(os.path.join(main_folder, 'web', 'static', 'data', 'annotations'))

  if not os.path.exists(os.path.join(main_folder, 'web', 'static', 'data', 'images')):
    os.makedirs(os.path.join(main_folder, 'web', 'static', 'data', 'images'))

  db = {'process_state': 'UNLABEL-PREPARE', 'round': 0}
  settings = {
    'port': server_port,
    'name': activelearning_name,
    'static_path': activelearning_server_static_dir,
    'template_path': activelearning_server_template_dir,
    'html_template': html_template,
    'keywords_template': keywords_template,
    'task_name': task.task_name,
    'task_type': task.task_type,
    'db': db,
    'data_folder': os.path.join(main_folder, 'web', 'static', 'data', 'annotations'),
    'cookie_secret': str(uuid.uuid4()),
    'request_queue': request_queue,
    'upload': upload_folder,
    'download': download_folder
  }
  app = tornado.web.Application(handlers=[(r"/", IndexHandler),
                                          (r"/activelearning/upload/", UploadHandler),
                                          (r"/activelearning/download/", DownloadHandler),
                                          (r"/activelearning/state/", ActiveLearningState)],
                                **settings)

  http_server = tornado.httpserver.HTTPServer(app)
  http_server.listen(options.port)

  tornado.ioloop.IOLoop.instance().start()


# activelearning_web_server('hello', '/Users/jian/Downloads/Aa', None, None, 10031, '')
